/* Copyright (c) 1994 by The XFree86 Project, Inc.*/
/* Copyright (c) 1998 Jakub Jelinek */
/* Copyright (c) 1997-2001 Red Hat, Inc. All rights reserved. */
/* Mercilessly hacked for use with Xorg by: */
/* Copyright (c) 2006 Gilbert Ashley <amigo@ibiblio.org> */

/* Permission is hereby granted, free of charge, to any person obtaining a */
/* copy of this software and associated documentation files (the \"Software\"),*/
/* to deal in the Software without restriction, including without limitation*/
/* the rights to use, copy, modify, merge, publish, distribute, sublicense,*/
/* and/or sell copies of the Software, and to permit persons to whom the*/
/* Software is furnished to do so, subject to the following conditions:*/

/* The above copyright notice and this permission notice shall be included in*/
/* all copies or substantial portions of the Software.*/

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR*/
/* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL*/
/* THE XFREE86 PROJECT BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,*/
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF*/
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE*/
/* SOFTWARE.*/

/* Except as contained in this notice, neither the name of the XFree86 */
/* Project or Red Hat, Inc. shall not be used in advertising or otherwise */
/* to promote the sale, use or other dealings in this Software without prior */
/* written authorization from the XFree86 Project.*/

/* teg@redhat.com - bugfixes, XFree 4 support, DDC probing               */   
/* pbrown@redhat.com - many bugfixes                                     */
/* jakub@redhat.com - added support for SBUS servers			 */
/* ewt@redhat.com - general cleanups -- I'm sure I added a bunch
		    of memory leaks throughout, but who cares?           */
/* msf@redhat.com - added kickstart/pci probing support                  */
/* acme@conectiva.com.br - i18n for ramdac_name, clockchip_name,         */
/*                         monitortype_name, more i18n strings,          */
/*                         translate_array macro                         */
/*                                                                       */
/* kickstart options can be given on commandline:                        */
/*  --kickstart             run in non-interactive mode                  */
/*  --card  <type>      type of video card ('Cards' file)                */
/*  --server  <type>    XF86_<type>                                      */
/*  --driver  <driver>  Driver for use in XFree 4                        */
/*  --monitor <type>    ASCII name of monitor, look in 'MonitorsDB'      */
/*  --hsync  <val>          Horizontal sync rate - monitortype overrides */
/*  --vsync  <val>          Vertical sync rate - monitortype overrides   */
/*  --bandwidth  <val>      Bandwidth of monitor - monitortype overrides */
/*  --preferxf3         Prefer XFree86 3.3 server if available           */
/*  --preferxf4         Prefer XFree86 4.0 server if available           */
/*                                                                       */
/* Usage:                                                                */
/*       Give --monitor    , OR give --hsync, --vsync, --bandwidth       */
/*       If any missing we just configure a Standard VGA device          */
/*       NOTE -> Need to log this somewhere                              */
/*                                                                       */
/*       --card used, if not defined then we pci-autoprobe               */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <newt.h>
#include <popt.h>
#include <locale.h>
#include <sys/ioctl.h>
#include <sys/vt.h>

#include "cards.h"
#include "combo.h"
#include <glib.h>
#include <kudzu/kudzu.h>

#include <kudzu/device.h>
#include <kudzu/ddc.h>

/* #include "pci-probing/pciprobe.h" */
/* #include "sbus-probing/sbusprobe.h" */
#include "XorgTest.h"

#ifdef __sparc__ 
#include <asm/kbio.h>
#endif

/* temporary debugging feature (yeah right) */
/* define both of these to enable, will write to DEBUG_XKB_FILE  */
/* #define DEBUG_XKB */
/* #define DEBUG_XKB_FILE "/tmp/XorgTest.log" */

struct monitorInfo {
    char * man, * name, * edid, * horizSync, * vertSync;
    int dpms;
} ;

struct monitorInfo *monitorInfo = NULL;
int monitorInfoNum = 0;
int dbUnavail = 0;

void modeline_to_res(char *m, int *x, int *y);

/*
 * Define the following to 310 to remove references to XFree86 features that
 * have been added since XFree86 3.1 (e.g. DoubleScan modes).
 */
#define XFREE86_VERSION 402

/*
 * Define this to have /etc/X11/xorg.conf prompted for as the default
 * location to write the XF86Config file to.
 */
#define PREFER_XORGIN_ETC


/*
 * Configuration variables.
 */

static char * progName;

int pick=0, rhcontinue=0, normal=0, custom=0, onboot=0, testing=0, expert=0;
int back = 0;
int noddcprobe=0;
int usefbdev=0;
int preferxf3=0,preferxf4=0;
int nodri=-1;
int probedmemory=0;
struct CardMonitorCombo* combo=NULL;

/* kickstart globals */
int kickstart=0;
char *ks_monitor=NULL;
char *ks_card=NULL;
char *ks_server=NULL;
char *ks_driver=NULL;
char *ks_hsync=NULL;
char *ks_vsync=NULL;
char *ks_bandwidth=NULL;
char *ks_mousedevice=NULL;
char *ks_mouse=NULL;
char *pciprobedcard=NULL;
int  ks_emulate3=0;

/* rest of the mess */

void get_out();
void cpfailed();
int screen_autoprobe(struct XconfigState *state);

/*
 * These are from the selected card definition. Parameters from the
 * definition are offered during the questioning about the video card.
 */

int card_selected;	/* Card selected from database. */
int card_screentype;
int card_accelserver;

char ** splitString(char * str, int length, char sep);

void write_XF86Config();
void write_XF86Config4();

void write_file();
char *read_file();

/* list of all servers, executable will have name 'XF86_<name>' */
static char *allservers[] = { "XF86_Mono", "XF86_VGA16", "XF86_SVGA", "XF86_S3", "XF86_Mach32",
			      "XF86_Mach8", "XF86_8514", "XF86_P9000", "XF86_AGX", "XF86_W32",
			      "XF86_Mach64", "XF86_I128", "XF86_S3V", "XF86_3DLabs",
#ifdef __alpha__
			      "XF86_TGA",
#endif
			      NULL };
#define NU_ALLSERVERS ((sizeof(allservers) / sizeof(*allservers)) - 1)

#ifdef __sparc__
#include <sys/utsname.h>
static int v9 = -1;

static int check_sunv9(void)
{
    if (v9 == -1) {
	struct utsname u;

        if (!uname (&u) && !strcmp (u.machine, "sparc64"))
            v9 = 1;
        else
            v9 = 0;
    }
    return v9;
}
#endif

#define NU_SUNSERVERS 3
static char *sunservers[] = {
	"xsunMono",
	"xsun",
	"xsun24",
	NULL
};

/*
 * This is the initial intro text that appears when the program is started.
 */

static char *intro_text =
N_("This program will create a basic xorg.conf file, based on menu selections you "
"make."
"\n\n"
"The xorg.conf file usually resides in /etc/X11. A sample "
"xorg.conf file is supplied with Xorg."
"\n\n"
"You can either take the sample xorg.conf as a base and edit it for your "
"configuration, or let this program produce a base xorg.conf file for your "
"configuration and fine-tune it. Refer to /usr/X11R6/lib/X11/doc/README.Config "
"for a detailed overview of the configuration process. "
"\n\n"
"For accelerated servers there are many chipset and card-specific options"
"and settings. This program does not know about these. On some configurations "
"some of these settings must be specified. Refer to the server man pages "
"and chipset-specific READMEs.");

static char *finalcomment_text =
N_("Configuration file has been written. Take a look at it before starting X. Note that "
"the xorg.conf file must be in one of the directories searched by the server "
"(e.g. /etc/X11/xorg.conf) in order to be used. While the server is running, pressing "
"ctrl, alt and '+' simultaneously cycles through the video resolutions. Pressing ctrl, "
"alt and backspace simultaneously immediately exits the server (use if "
"the monitor doesn't sync for a particular mode). "
"\n\n"
"For further configuration, refer to /usr/X11R6/lib/X11/doc/README.Config.\n"
"\n");

int custommonitor_configuration(struct XconfigState *state, int dir);

int customsyncrate_configuration(struct XconfigState *state);

int answerisyes(s)
	char *s;
{
	if (s[0] == '\'')	/* For fools that type the ' literally. */
		return tolower(s[1]) == 'y';
	return tolower(s[0]) == 'y';
}

/*
 * For taking backups of XF86Config{,-4} files
 */

int unconditionalCopy(char *filename1, char *filename2)
{
    FILE *in,*out;
    char *data;
    struct stat statbuf;
    
    if(stat(filename2,&statbuf)==0){
        if(!unlink(filename2))
            return -1;
    }
    if((in=fopen(filename1,"r"))==NULL)
        return -1;
    if((out=fopen(filename2,"w"))==NULL)
        return -1;
    if((stat(filename1,&statbuf))<0)
        return -1;
    data=malloc(statbuf.st_size);
    fread(data,1,statbuf.st_size,in);
    fwrite(data,1,statbuf.st_size,out);
    free(data);
    fclose(in);
    fclose(out);
    return 0;
}

void backupXConfigs()
{
    struct stat statbuf;

    if(stat("/etc/X11/xorg.conf",&statbuf)==0){
        unconditionalCopy("/etc/X11/xorg.conf","/etc/X11/xorg.conf.old");
    }
    /* if(stat("/etc/X11/XF86Config-4",&statbuf)==0){
        unconditionalCopy("/etc/X11/XF86Config-4","/etc/X11/XF86Config-4.old");
    } */
}

int rewriteInittab(struct XconfigState *state)
{
    FILE *inittab;
    FILE *newInittab;
    char *s;
    char buf[250];

    if (testing) return 0;

    inittab = fopen("/etc/inittab", "r");
    if (!inittab) {
	fprintf(stderr, _("%s: cannot open /etc/inittab: %s\n"),
		progName, strerror(errno));
	return 1;
    }
    newInittab = fopen("/etc/inittab-", "w");
    if (!newInittab) {
	fclose(inittab);
	fprintf(stderr, _("%s: cannot create /etc/inittab-: %s\n"),
		progName, strerror(errno));
	return 1;
    }
    while ((fgets(buf, sizeof(buf) - 1, inittab)) != NULL) {
	s = buf;

	/* leading spaces */
	while (*s && isspace(*s))
	    s++;

	if (strlen(s) >= 2) {
	    if (!strncmp(s, "id:", 2)) {
		/* found the line we're looking for, rewrite it */
		fprintf(newInittab, "id:%d:initdefault:\n", state->runLevel);
		continue;
	    }
	}
	/* everything else falls through here */
	fputs(buf, newInittab);
    }
    fclose(newInittab);
    fclose(inittab);

    unlink("/etc/inittab");
    rename("/etc/inittab-", "/etc/inittab");

    return 0;
}

static void startXfs(int normal)
{
  struct stat sb;

  /* if normal = 0, we are running under installer.  Bail. */
  if (!normal)
    return;

  /* 
     if there is no pidof program, or chkconfig, or the initscripts
     functions script, or /proc is not mounted, we need to bail.
  */
  if (stat("/sbin/pidof", &sb))
    return;
  if (stat("/sbin/chkconfig", &sb))
    return;
  if (stat("/etc/rc.d/init.d/functions", &sb))
    return;
  if (stat("/proc/mounts", &sb))
    return;

  system("/sbin/chkconfig --add xfs");
  if (system("source /etc/rc.d/init.d/functions; status xfs >/dev/null")) {
      system("/etc/rc.d/init.d/xfs restart >/dev/null");
  }
}

static int slow_X(void)
{
#ifdef __sparc__
  /* Some old SPARC boxes are too slow for the 10 seconds timeout */
  FILE *f = fopen("/proc/cpuinfo","r");
  char buf[1024], *p;
  if (f) {
    while (fgets (buf, 1024, f) != NULL)
      if (!strncmp (buf, "type", 4)) {
	for (p = buf + 4; *p == ' ' || *p == '\t' || *p == ':'; p++);
	if (!strncmp (p, "sun4", 4)) {
	  /* sun4c and sun4 are very slow */
	  if (p[4] == 'c' || p[4] == '\n')
	    return 1;
	}
      }
  }
#endif
  return 0;
}

static void setupModeRow(newtGrid grid, newtComponent * subform,
			 int startCol, int startRow,
			 char * modes, char * boxes, int padding) {
    char ** list;
    newtComponent box;
    int i;

    if (!modes) {
	*subform = NULL;
	return;
    }
    list = splitString(modes, strlen(modes), ' ');
    *subform = newtForm(NULL, NULL, 0);
    i = 0;
    while (list[i] != NULL) {
	/* don't _ever_ use foo(list[i++], bar+i); !!! */
	box = newtCheckbox(-1, -1, list[i], ' ', NULL, boxes + i);
	i++;
	newtFormAddComponent(*subform, box);
	newtGridSetField(grid, startCol, startRow, NEWT_GRID_COMPONENT, box,
			 padding, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	startRow++;
    }

    free(list);
}

void rebuildModes(char ** modes, char * boxes) {
    char ** list;
    int i;
    char buf[100];

    *buf = '\0';

    if (!*modes) {
	return;
    }

    list = splitString(*modes, strlen(*modes), ' ');
    i = 0;
    while (list[i] != NULL) {
	if (boxes[i] != ' ') {
	    if (buf[0]) {
		strcat(buf, " ");
	    }
	    strcat(buf, list[i]);
	}
	i++;
    }

    if (strlen(buf))
	*modes = strdup(buf);
    else
	*modes = NULL;

    free(list);
}

void findMaxRes(char * modes, int * maxx, int * maxy) {
    char ** list;
    int i, x, y;

    *maxx = *maxy = 0;

    if (!modes) {
	return;
    }
    list = splitString(modes, strlen(modes), ' ');
    i = 0;
    while (list[i] != NULL) {
	modeline_to_res(list[i], &x, &y);
	if (x > *maxx) *maxx = x;
	if (y > *maxy) *maxy = y;
	i++;
    }

    free(list);
}

int chooseXmodes(struct XconfigState *state, char ** modes8ptr, char ** modes16ptr, char ** modes32ptr) {
    newtComponent textbox, form, subform[3], ok, cancel, answer;
    newtGrid subgrid, grid, buttons;
    char selModes1[6];
    char selModes2[6];
    char selModes3[6];

    textbox = newtTextboxReflowed(-1, -1,
		       _("Select the video modes you would like to use.  "
		         "8 bit modes allow for 256 colors, 16 bit modes "
		         "allow for 64k colors, and 24 bit modes allow for "
		         "true color.  Performance will be slower, however, "
		         "the higher you go.  You should select at least one "
		         "of the elements below."), 56, 0, 20, 0);

    form = newtForm(NULL, NULL, 0);

   /* be sure that the y in newtCreateGrid(x, y) is tall enough
      for n resolutions + 1 */
    subgrid = newtCreateGrid(3, 8);
    newtGridSetField(subgrid, 0, 0, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("8 bit:")),
		     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
    newtGridSetField(subgrid, 1, 0, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("16 bit:")),
		     4, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
    newtGridSetField(subgrid, 2, 0, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("24 bit:")),
		     4, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridAddComponentsToForm(subgrid, form, 1);

    setupModeRow(subgrid, &subform[0], 0, 1, *modes8ptr, selModes1, 0);
    setupModeRow(subgrid, &subform[1], 1, 1, *modes16ptr, selModes2, 4);
    setupModeRow(subgrid, &subform[2], 2, 1, *modes32ptr, selModes3, 4);

    buttons = newtButtonBar(_("Ok"), &ok, _("Back"), &cancel, NULL);
    grid = newtGridBasicWindow(textbox, subgrid, buttons);

    newtGridWrappedWindow(grid, _("Select Video Modes"));
    newtFormAddComponent(form, textbox);
    if (subform[0]) newtFormAddComponent(form, subform[0]);
    if (subform[1]) newtFormAddComponent(form, subform[1]);
    if (subform[2]) newtFormAddComponent(form, subform[2]);
    newtGridAddComponentsToForm(buttons, form, 1);
    newtGridFree(grid, 1);

    answer = newtRunForm(form);
    newtFormDestroy(form);
    newtPopWindow();

    if (answer == cancel) {
      return 1;
    }

    rebuildModes(modes8ptr, selModes1);
    rebuildModes(modes16ptr, selModes2);
    rebuildModes(modes32ptr, selModes3);

    return 0;
}

/*
 * Mouse configuration. Read /etc/sysconfig/mouse and set globals vars
 */

static int readMouseConfigFile(char ** mouse, char ** emu, char **mouseDev) {
    FILE * f;
    char * start, * end;
    char buf[250];
    int line = 0;

    *mouseDev = NULL;
    
    f = fopen("/etc/sysconfig/mouse", "r");
    if (!f) {
        if (errno == ENOENT) {
            return 1;
        }

	if (!kickstart)
	    newtFinished();
        fprintf(stderr, _("%s: cannot open /etc/sysconfig/mouse: %s\n"),
                progName, strerror(errno));
        exit(1);
    }

    errno = 0;
    while (fgets(buf, sizeof(buf) - 1, f)) {
        line++;

        start = buf;

        /* skipping leading spaces */
        while (*start && isspace(*start)) start++;
        if (!*start) continue;

        /* skip comments */
        if (*start == '#') continue;

        /* cut off trailing spaces and \n */
        end = start + strlen(start) - 2;
        while (isspace(*end)) end--;
        end++;
        *end = '\0';
        end--;

        if (!strncmp("MOUSETYPE=", start, 10)) {
            start += 10;
            if (*start == '"' && *end != '"') {
                fprintf(stderr, _("%s: mismatched quotes on line %d in "
                        "/etc/sysconfig/mouse\n"), progName, line);
                fclose(f);
                return 1;
            } else if (*start == '"') {
                start++;
                *end = '\0';
                end--;
            }

            *mouse = strdup(start);
        }
	else if (!strncmp("XMOUSETYPE=", start, 11)) {
	  start += 11;
	  if (*start == '"' && *end != '"') {
	    fprintf(stderr, _("%s: mismatched quotes on line %d in "
			      "/etc/sysconfig/mouse\n"), progName, line);
	    fclose(f);
	    return 1;
	  } else if (*start == '"') {
	    start++;
	    *end = '\0';
	    end--;
	  }

	  *mouse = strdup(start);
        }
        else if (!strncmp("XEMU3=", start, 6)) {
            start += 6;
            if (*start == '"' && *end != '"') {
                fprintf(stderr, _("%s: mismatched quotes on line %d in "
                        "/etc/sysconfig/mouse\n"), progName, line);
                fclose(f);
                return 1;
            } else if (*start == '"') {
                start++;
                *end = '\0';
                end--;
            }

            *emu = strdup(start);
        }
	else if (!strncmp("DEVICE=", start, 7)) {
            start += 7;
            if (*start == '"' && *end != '"') {
                fprintf(stderr, _("%s: mismatched quotes on line %d in "
                        "/etc/sysconfig/mouse\n"), progName, line);
                fclose(f);
                return 1;
            } else if (*start == '"') {
                start++;
                *end = '\0';
                end--;
            }

            *mouseDev = strdup(start);
	}
	else if (!strncmp("FULLNAME=", start, 9)) {
	  /* do nothing if we see this line */
	  ;
	}

        else {
            fprintf(stderr, _("%s: line %d unexpected in "
		    "/etc/sysconfig/mouse\n"), progName, line);
            fclose(f);
            return 1;
        }
	errno = 0;
    }
    if (errno) {
	if (!kickstart)
	    newtFinished();
        fprintf(stderr, _("%s: cannot read /etc/sysconfig/mouse: %s\n"),
                progName, strerror(errno));
    }

    if (*mouseDev == NULL) /* default */
	*mouseDev = strdup("/dev/mouse");
    
    return 0;
}


/*                                                          */
/* run mouseconfig, see if they properly configured a mouse */
/*                                                          */
void mouse_configuration(struct XconfigState *state) {
  int childpid;
  int status;
  pid_t wpid;
  int argc;
  char * mouse, * emu;
  char *args[5]={"/usr/sbin/mouseconfig",NULL,NULL,NULL,NULL};

  status = readMouseConfigFile( &mouse, &emu, &state->pointerdevice );
  if (status) {
      argc=1;
      if (!kickstart) {
	  newtSuspend();
      } else {
	  if (ks_mousedevice) {
	      args[argc++] = "--device";
	      args[argc++] = ks_mousedevice;
	  }

	  if (ks_emulate3)
	      args[argc++] = "--emulthree";

	  if (ks_mouse)
	      args[argc++] = ks_mouse;
	  else
	      args[argc++] = "microsoft";
      }

      if (!(childpid = fork())) {
	  execv("/usr/sbin/mouseconfig", args);
	  exit(2);
      }

      wpid = waitpid(childpid, &status, 0);

      if (!kickstart)
	  newtResume();
      status = readMouseConfigFile( &mouse, &emu, &state->pointerdevice );
      if (status) {
	  if (!kickstart)
	      newtFinished();
	  fprintf(stderr, _("%s: Critical error reading "
		 "/etc/sysconfig/mouse.\n"), progName);
	  fprintf(stderr, _("%s: Check permissions of "
		 "/etc/sysconfig/mouse.\n"), progName);
	  fprintf(stderr, _("%s: Also - make sure you are running this "
		   "program as root\n"), progName);
	  exit(1);
      }
  }

  strncpy( state->mousetype, mouse, sizeof(state->mousetype) );
  state->chordmiddle = 0;
  if (answerisyes(emu))
      state->emulate3buttons = 1;
  else
      state->emulate3buttons = 0;
}


/*
 * Keyboard configuration.
 */

/* reads /etc/sysconfig/keyboard, which just has KEYTABLE=<path> */
/* double quotes are stripped, #'s denote comments */
static int readKBDConfigFile(char ** table) {
    FILE * f;
    char * start, * end;
    char buf[250];
    int line = 0;

    f = fopen("/etc/sysconfig/keyboard", "r");
    if (!f) {
	if (errno == ENOENT) {
	    return 1;
	}
	if (!kickstart)
	    newtFinished();
	fprintf(stderr, _("%s: cannot open /etc/sysconfig/keyboard: %s\n"),
		progName, strerror(errno));
	exit(1);
    }

    /* we got here ok, so lets zero errno? */
    /* otherwise broken code below incorrectly */
    /* detects a failure condition */
    errno = 0;

    while (fgets(buf, sizeof(buf) - 1, f)) {
	line++;

	start = buf;

	/* skipping leading spaces */
	while (*start && isspace(*start)) start++;
	if (!*start) continue;

	/* skip comments */
	if (*start == '#') continue;

	/* cut off trailing spaces and \n */
	end = start + strlen(start) - 2;
	while (isspace(*end)) end--;
	end++;
	*end = '\0';
	end--;

	if (!strncmp("KEYTABLE=", start, 9)) {
	    start += 9;
	    if (*start == '"' && *end != '"') {
		fprintf(stderr, _("%s: mismatched quotes on line %d in "
			"/etc/sysconfig/keyboard\n"), progName, line);
		fclose(f);
		return 1;
	    } else if (*start == '"') {
		start++;
		*end = '\0';
		end--;
	    }

	    if (*end != 'p' || *(end - 1) != 'a' || *(end - 2) != 'm' ||
		*(end - 3) != '.') {
	      *table = strdup(start);
	    } else {
               end -= 3;
               *end = '\0';

               end = start;
               while (*end) {
                   if (*end == '/')
                       start = end + 1;
                   end++;
               }
               *table = strdup(start);
           }
	} else if (strncmp("KEYBOARDTYPE=", start, 13)) {
	    fprintf(stderr, _("%s: line %d unexpected in "
		    "/etc/sysconfig/keyboard\n"), progName, line);
	    fclose(f);
	    return 1;
	}
    }

    if (errno) {
	if (!kickstart)
	    newtFinished();
	fprintf(stderr, _("%s: cannot read /etc/sysconfig/keyboard: %s\n"),
		progName, strerror(errno));
    }

    return 0;
}

/* it is assumed if a keymap DOES NOT appear in this translation table */
/* this it maps to us.map                                              */
/* maps from linux keymap name to the XKBD layout type                 */
#define KEYMAP_TRANSLATE_NUM ((sizeof(keymap_translate))/(sizeof(char *)))

static char *keymap_translate[] = {
  "be-latin1 -> be", "be2-latin1 -> be", "bg -> bg", "cf -> ca_enhanced",
  "de-latin1-nodeadkeys -> de", "de-latin1 -> de", "de -> de",
  "dk-latin1 -> dk", "dk -> dk", "dvorak -> dvorak", "es -> es",
  "fi-latin1 -> fi", "fi -> fi", "fr-latin1 -> fr", "fr -> fr",
  "fr_CH-latin1 -> fr_CH", "hu -> hu",
  "hu101 -> hu", "it-ibm -> it", "it -> it", "it2 -> it", "is-latin1 -> is",
  "is -> is", "no-latin1 -> no", "no -> no", "pl -> pl", "ru -> ru",
  "ru1 -> ru", "ru2 -> ru", "no -> no", "pl -> pl", "ru -> ru", "ru1 -> ru",
  "ru2 -> ru", "russian -> ru", "se-latin1 -> se", "sg-latin1 -> de_CH",
  "uk -> gb", "jp106 -> jp"
};

#ifdef __sparc__
#define SUNKEYMAP_TRANSLATE_NUM ((sizeof(sunkeymap_translate))/(sizeof(char *)))
static char *sunkeymap_translate[] = {
  "sunt5-de-latin1 -> de", "sunt4-es -> es", "sunt5-es -> es", "sunt5-uk -> gb",
  "sunt5-us-cz -> cs", "sunt5-cz-us -> cs"
    };
#endif

void keyboard_configuration(struct XconfigState *state) {
    char linux_keymap[40], trans_keymap[40];
    char *table = NULL;
    int i, found;
    int childpid, status;

    /* this should be incorporated into a dialog box */
    state->altmeta = 0;

    table = "us";
    status = readKBDConfigFile(&table);
    if (status && !kickstart) {
	newtSuspend();
	if (!(childpid = fork())) {
	    execl("/usr/sbin/kbdconfig", "", NULL);
	    exit(2);
	}
	waitpid(childpid, &status, 0);
	if (!kickstart)
	    newtResume();

	status = readKBDConfigFile(&table);
	if (status) {
	    if (!kickstart)
		newtFinished();
	    fprintf(stderr, _("%s: Critical error reading "
		    "/etc/sysconfig/keyboard.\n"), progName);
	    fprintf(stderr, _("%s: Check permissions of "
		    "/etc/sysconfig/keyboard.\n"), progName);
	    fprintf(stderr, _("%s: Also - make sure you are running "
		    "this program as root\n"), progName);
	    exit(1);
	}
    }

    if (!table)
	table = "us";

#ifdef __sparc__
    {
	struct device **devs;
	char *p, *q;

	devs = probeDevices(CLASS_KEYBOARD,BUS_KEYBOARD,PROBE_ONE);
	if (!devs || !strncmp (devs[0]->desc, "Serial console", 14)) {
	    if (!kickstart)
		newtFinished();
	    fprintf (stderr, _("%s: Cannot find any keyboards\n"), progName);
	    exit(1);
	}
	
	if (!strncmp (devs[0]->desc, "Sun Type", 8)) {
	    state->xkb_rules = "sun";
	    state->xkb_model = "type5";
	    state->xkb_keycodes = "sun(type5)";
	    state->xkb_geometry = "sun";
	    state->xkb_symbols = "sun/us(sun5)";
	    state->xkb_layout = "us";
	    p = devs[0]->desc + 8;
	    if (*p == '4') {
		state->xkb_model = "type4";
		state->xkb_keycodes = "sun(type4)";
		state->xkb_geometry = "sun(type4)";
		state->xkb_symbols = "sun/us(sun4)";
	    } else if (!strncmp (p, "5 Euro", 6)) {
		state->xkb_model = "type5_euro";
		state->xkb_geometry = "sun(type5euro)";
		p += 5;
	    } else if (!strncmp (p, "5 Unix", 6)) {
		/* state->xkb_model = "type5_unix"; */
		state->xkb_geometry = "sun(type5unix)";
		p += 5;
	    }
	    p += 2;
	    for (i = 0, found = 0; i < SUNKEYMAP_TRANSLATE_NUM; i++) {
		sscanf(keymap_translate[i],"%s -> %s\n", linux_keymap, trans_keymap);
		if (!strncmp(table, linux_keymap, 40)) {
		    found = 1;
		    break;
		}
	    }
	    if (!found && strncmp (p, "Keyboard", 8) && strncmp (p, "nl", 2) &&
	        strncmp (p, "ko", 2) && strncmp (p, "tw", 2)) {
		if (!strncmp (p, "fr_BE", 5))
		    strcpy (trans_keymap, "be");
		else if (!strncmp (p, "fr_CA", 5))
		    strcpy (trans_keymap, "fr");
		else {
		    q = strchr (p, ' ');
		    memcpy (trans_keymap, p, q - p);
		    trans_keymap [q - p] = 0;
		}
		found = 1;
	    }
	    if (found) {
		p = malloc (strlen (state->xkb_symbols) + strlen (trans_keymap) + 4);
		if (p) {
		    sprintf (p, "%s + %s", state->xkb_symbols, trans_keymap);
		    state->xkb_symbols = p;
		}
	    }
	    return;
	} else if (state->screentype == 5) {
	    if (!kickstart)
		newtFinished();
	    fprintf (stderr, _("%s: xsun* servers don't support non-Sun keyboards\n"), progName);
	    exit(1);
	}

	devs[0]->freeDevice(devs[0]);
	free(devs);
    }

#endif /* __sparc__ */

    for (i = 0, found = 0; i < KEYMAP_TRANSLATE_NUM; i++) {
	sscanf(keymap_translate[i],"%s -> %s\n", linux_keymap, trans_keymap);
	if (!strncmp(table, linux_keymap, 40)) {
	    found = 1;
	    break;
	}
    }

    if (!found)
	state->xkb_keymap = "us";
    else
	state->xkb_keymap = strdup(trans_keymap);
}

/*
 * Monitor configuration.
 */

static char *monitorintro_text =
N_("Now we want to set the specifications of the monitor. The two critical "
"parameters are the vertical refresh rate, which is the rate at which the "
"the whole screen is refreshed, and most importantly the horizontal sync rate, "
"which is the rate at which scanlines are displayed."
"\n\n"
"The valid range for horizontal sync and vertical sync should be documented "
"in the manual of your monitor.");

static char *hsyncintro_text =
N_("You must indicate the horizontal sync range of your monitor. You can either "
"select one of the predefined ranges below that correspond to industry-"
"standard monitor types, or give a specific range."
"\n"
"It is VERY IMPORTANT that you do not specify a monitor type with a horizontal "
"sync range that is beyond the capabilities of your monitor.");

static char *vsyncintro_text =
N_("You must indicate the vertical sync range of your monitor. You can either "
"select one of the predefined ranges below that correspond to industry-"
"standard monitor types, or give a specific range. For interlaced modes, "
"the number that counts is the high one (e.g. 87 Hz rather than 43 Hz).");

static char *vsync_range[] = { "50-70 Hz", "50-90 Hz", "50-100 Hz", "40-150 Hz", NULL };
#define NU_VSYNCTANGE ((sizeof(vsync_range) /  \
				sizeof(*vsync_range)) - 1)

static char *monitortype_range[] = {
        "Custom",
	"31.5",
	"31.5 - 35.1",
	"31.5, 35.15, 35.5",
	"31.5 - 37.9",
	"31.5 - 48.5",
	"31.5 - 57.0",
	"31.5 - 64.3",
	"31.5 - 79.0",
	"31.5 - 82.0",
	"31.5 - 88.0",
	"31.5 - 94.0",
	"31.5, 35.5",
	NULL
};
#define NU_MONITORTYPES ((sizeof(monitortype_range) /  \
				sizeof(*monitortype_range)) - 1)

static char *monitortype_name[NU_MONITORTYPES + 1] = {
        N_("Custom"),
	N_("Standard VGA, 640x480 @ 60 Hz >31.5"),
	N_("Super VGA, 800x600 @ 56 Hz  31.5 >35.1"),
	N_("Super VGA, 1024x768 @ 87 Hz interlaced >31.5, 35.15, 35.5"),
	N_("Extended Super VGA, 800x600 @ 60 Hz >31.5 - 37.9"),
	N_("Non-Interlaced SVGA, 1024x768 @ 60 Hz >31.5 - 48.5"),
	N_("High Frequency SVGA, 1024x768 @ 70 Hz >31.5 - 57.0"),
	N_("Monitor that can do 1280x1024 @ 60 Hz >31.5 - 64.3"),
	N_("Monitor that can do 1280x1024 @ 74 Hz >31.5 - 79.0"),
	N_("Monitor that can do 1280x1024 @ 76 Hz >31.5 - 82.0"),
	N_("Monitor that can do 1600x1200 @ 70 Hz >31.5 - 88.0"),
	N_("Monitor that can do 1600x1200 @ 76 Hz >31.5 - 94.0"),
	N_("8514 Compatible, 1024x768 @ 87 Hz interlaced >31.5, 35.5"),
	NULL
};

static char *monitorabbrev_name[NU_MONITORTYPES + 1] = {
	"640x480@60",
	"800x600@56",
	"1024x768@87i",
	"800x600@60",
	"1024x768@60",
	"1024x768@70",
	"1280x1024@60",
	"1280x1024@74",
	"1280x1024@76",
	"1600x1200@70",
	"1600x1200@76",
	"8514 compatible",
	NULL
};

/* Read monitor data in from disk */
int ReadMonitorDB( void ) {
    FILE *mondbfile;
    char line[256];
    char *s, *t;
    char *man, *name, *edid, *horsync, *versync;
    int dpms = 1; /* optional field */
    int num = 0;

    if (NULL == (mondbfile = fopen(MONITOR_DATABASE_FILE, "r")) )
		return -1;

	while(1) {
    if (!fgets(line, sizeof(line), mondbfile))
        break;

	/* trim initial spaces, continue if line is blank or comment */
	for (s=line; *s && isspace(*s); s++);

	if (!*s || *s == '#')
	    continue;

	/* find first field delim (manufacturer) */
	for (t=s; *t && *t != ';'; t++);

	if (!*t)
	    continue;

	*t = 0;
	man = s;
	s = t+1;

	/* find second field delim (model) */
	for (; *s && isspace(*s); s++);
	for (t=s; *t && *t != ';'; t++);

	if (!*t)
	    continue;

	*t = 0;
	name = s;
	s = t+1;

	/* find third field delim (EDID) */
	for (; *s && isspace(*s); s++);
	for (t=s; *t && *t != ';'; t++);

	if (!*t)
	    continue;

	*t = 0;
	edid = s;
	s = t+1;

	/* find fourth field delim (horizontal refresh range) */
	for (; *s && isspace(*s); s++);
	for (t=s; *t && *t != ';'; t++);

	if (!*t)
	    continue;

	*t = 0;
	horsync = s;
	s = t+1;

	/* find fourth field delim (vertical refresh range) */
	for (; *s && isspace(*s); s++);
	for (t=s; *t && *t != ';' && *t != '\n'; t++);
	*t = 0;
	versync = s;

	/* find fifth field delim (dpms) */
	for (; *s && isspace(*s); s++);
	for (t=s; *t && *t != ';' && *t != '\n'; t++);
	t = 0;
	if (s[0] == '0')
	    dpms = 0;
	else
	    dpms = 1;
	    
	/* make an entry */
	if (num == 0)
	    monitorInfo = (struct monitorInfo *) malloc(sizeof(*monitorInfo));
	else
	    monitorInfo=(struct monitorInfo *)realloc(monitorInfo,
					     (num+1)*sizeof(*monitorInfo));

	monitorInfo[num].man       = strdup(man);
	monitorInfo[num].name      = strdup(name);
	monitorInfo[num].edid      = strdup(edid);
	monitorInfo[num].horizSync = strdup(horsync);
	monitorInfo[num].vertSync  = strdup(versync);
	monitorInfo[num].dpms      = dpms;
	num++;
    }

    /* add null entry to mark end of list */
    monitorInfo=(struct monitorInfo *)realloc(monitorInfo,
					      (num+1)*sizeof(*monitorInfo));
    memset(&monitorInfo[num], 0, sizeof(*monitorInfo));
    monitorInfoNum = num;
    fclose(mondbfile);
    return 0;
}

void ListMonitorTypes(void) {
    int i;

    if (ReadMonitorDB()) {
	fprintf(stderr, _("\n\nMonitor database was not found!\n\n"));
	return;
    }

    for (i=0; i < monitorInfoNum; i++)
	fprintf(stdout,"                        %s\n",monitorInfo[i].name);

    fprintf(stderr,
	_("\n      These generic types are also supported (use only the 'Generic##' part\n"
	"      when specifying the monitor). You must specify the preceding\n"
	"      '0' if the type is less than 10 (ie. 'Generic02')\n\n"));

    for (i=0; i<NU_MONITORTYPES; i++)
	fprintf(stdout,"        Generic%02d - "
		"%s\n",i,monitortype_name[i]);

}

/* return 1 on cancel, 2 on error, 0 on success */

int monitor_configuration(struct XconfigState *state) {
    int i, j, cols, rows;
    int rc, res = 0;
    char ** monitors;
    char *monident, *monhrange, *monvrange;
    char tmpstr[80];
    newtComponent form, ok, no, answer, textbox, textbox2;
    newtGrid grid, buttons, subgrid;
#ifdef __i386__
    struct device **monitordevs;
    struct ddcDevice *ddcmonitor;
#endif

    /* read monitor database */
    if (monitorInfoNum == 0)
	dbUnavail = ReadMonitorDB();
    
    /* temp fix - try to probe monitor, report results */
    monident  = NULL;
    monhrange = NULL;
    monvrange = NULL;

#ifdef __i386__
	if (noddcprobe == 0) {
        initializeBusDeviceList(BUS_DDC);
        monitordevs = probeDevices(CLASS_MONITOR, BUS_DDC, PROBE_ALL);
        if (monitordevs) {
            /* Don't even try to handle more than one monitor */
            ddcmonitor=(struct ddcDevice*) monitordevs[0]; 
            if((ddcmonitor->desc)!=NULL){
                monident=strdup(ddcmonitor->desc);
            } else if((ddcmonitor->id)!=NULL) {
                monident=strdup(ddcmonitor->id);
		g_strup(monident);
            } else {
                monident=strdup(_("Unknown monitor"));
            }
            probedmemory = (int) (ddcmonitor->mem);
	    
            monhrange=malloc(100);
            monvrange=malloc(100);
            snprintf(monhrange,100,"%i - %i",ddcmonitor->horizSyncMin,
                     ddcmonitor->horizSyncMax);
            snprintf(monvrange,100,"%i - %i",ddcmonitor->vertRefreshMin,
                     ddcmonitor->vertRefreshMax);
            
            if((ddcmonitor->horizSyncMin==0)&&(ddcmonitor->horizSyncMax==0)
               &&(ddcmonitor->vertRefreshMin==0)&&(ddcmonitor->vertRefreshMax==0)){
		/*
		 * Try to look up refresh rates in the monitor database
		 * when DDC probing for these values fails.
		 */
		for (i = 0; monitorInfo && monitorInfo[i].name; i++) {
		    /*
		     * convert to upper case.  monident already is
		     * upper-case.
		     */
		    g_strup(monitorInfo[i].edid);
		    if (strcmp(monident, monitorInfo[i].edid) == 0) {
			res = 1;
			break;
		    }
		}
		if (res) {
		    /* use refresh rates from the file. */
		    strncpy(monhrange, monitorInfo[i].horizSync, 100);
		    strncpy(monvrange, monitorInfo[i].vertSync, 100);
		    /* change the name as well. */
		    free(monident);
		    monident = strdup(monitorInfo[i].name);
		} else {
		    free(monident);
		    monident=NULL;		    
		}
		
            } /* if bad refresh rates */
            for (i=0; monitordevs[i] ; i++)
                monitordevs[i]->freeDevice(monitordevs[i]);
            free(monitordevs);
        } /* DDC successful */
    } /* probe ddc? */
#endif
    
    if (monident) {
	if (!kickstart){
	    buttons = newtButtonBar(_("Yes"), &ok, _("No"), &no, NULL);
	    textbox = newtTextboxReflowed(-1, -1, 
					  _("Monitor probing found:"), 
					  50, 10, 20, 0);
	    textbox2 = newtTextboxReflowed(-1, -1, 
					  _("Do you want to use these settings?"), 
					  50, 10, 20, 0);
	    
	    form = newtForm(NULL, NULL, 0);
	    
	    subgrid = newtCreateGrid(2, 3);
	    
	    newtGridSetField(subgrid, 0, 0, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1, _("Monitor Name         :")),
			     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    newtGridSetField(subgrid, 0, 1, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1, _("Horizontal Sync Range:")),
			     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    newtGridSetField(subgrid, 0, 2, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1, _("Vertical Sync Range  :")),
			     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    newtGridSetField(subgrid, 1, 0, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1, monident),
			     2, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

	    snprintf(tmpstr, sizeof(tmpstr), "%s kHz", monhrange);
	    newtGridSetField(subgrid, 1, 1, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1,tmpstr),
			     2, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    
	    snprintf(tmpstr, sizeof(tmpstr), "%s Hz", monvrange);
	    newtGridSetField(subgrid, 1, 2, NEWT_GRID_COMPONENT,
			     newtLabel(-1, -1, tmpstr),
			     2, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    
	    newtGridAddComponentsToForm(subgrid, form, 1);
	    
	    grid = newtCreateGrid(1, 4);
	    newtGridSetField(grid, 0, 0, NEWT_GRID_COMPONENT, textbox,
			     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
	    newtGridSetField(grid, 0, 1, NEWT_GRID_SUBGRID, subgrid,
			     0, 1, 0, 0, 0, 0);
	    newtGridSetField(grid, 0, 2, NEWT_GRID_COMPONENT, textbox2,
			     0, 1, 0, 0, 0, NEWT_GRID_FLAG_GROWX);
	    newtGridSetField(grid, 0, 3, NEWT_GRID_SUBGRID, buttons,
			     0, 1, 0, 0, 0, NEWT_GRID_FLAG_GROWX);

	    newtGridWrappedWindow(grid, "Monitor Probe");
	    newtFormAddComponent(form, textbox);
	    newtGridAddComponentsToForm(buttons, form, 1);
	    newtGridFree(grid, 1);
	    
	    newtFormAddComponent(form, textbox2);
	    answer = newtRunForm(form);
	    if (answer != no)
		rc = 1;
	    else
		rc = 2;
	    newtPopWindow();
	    newtFormDestroy(form);
	} else {
	    /* if kickstart file specifies a monitor, use it instead */
	    /* to allow option of overriding probed values           */
	    if (!ks_monitor)
		rc = 1;
	    else 
		rc = 2;
	}
	    
	if (rc != 2) {
	    state->hsyncrange = monhrange;
	    state->vsyncrange = monvrange;
	    state->dpms = 1;
	    state->monitoridentifier = monident;
	    state->monitorvendorname = "Unknown";
	    state->monitormodelname = "Unknown";
	    return 0;
	}
    }

    if (dbUnavail && !kickstart) {
	newtWinMessage(_("Monitor DB"), _("Ok"),
		       _("The monitor database file %s could "
			 "not be loaded. Please select a "
			 "custom monitor configuration closest "
			 "to your actual monitor."),
		       MONITOR_DATABASE_FILE);
	
	return custommonitor_configuration(state, 1);
    }

    
    /* see if we're kickstarting, user had to supply monitor type */
    if (kickstart) {
	int sel=-1;
	if (ks_monitor) {

	    /* see if they gave a generic type first */
	    if (!strncasecmp(ks_monitor,"generic",7)) {
		int idx;

		if (ks_monitor[7] && isdigit(ks_monitor[7]) &&
		    ks_monitor[8] && isdigit(ks_monitor[8])) {
		    idx = (ks_monitor[7]-48)*10+(ks_monitor[8]-48);

		    if (idx >= 0 && idx < NU_MONITORTYPES) {
			state->hsyncrange = monitortype_range[idx];
			if (idx < 4)
			    /* use 60 hz for low handwidth monitors */
			    state->vsyncrange = "50-61";
			else if (idx < 7)
			    state->vsyncrange = "50-70";
			else
			    state->vsyncrange = "50-90";
			state->dpms = 1;
			state->monitoridentifier = monitortype_name[idx];
			state->monitorvendorname = "Unknown";
			state->monitormodelname = "Unknown";
			return 0;
		    }
		}
	    }

	    /* didnt find a valid generic monitor, keep trying */
	    for (i = 0; monitorInfo && monitorInfo[i].name; i++)
		if (!strcasecmp(ks_monitor, monitorInfo[i].name)) {
		    sel = i;
		    break;
		}

	    if (sel != -1) {
		state->hsyncrange = monitorInfo[sel].horizSync;
		state->vsyncrange = monitorInfo[sel].vertSync;
		state->dpms = 1;
		state->monitoridentifier = monitorInfo[sel].name;
		state->monitorvendorname = "Unknown";
		state->monitormodelname = "Unknown";
		return 0;
	    } else {
		for (i = 0; i < NU_MONITORTYPES; i++)
		    if (!strcasecmp(ks_monitor,  monitortype_name[i]) ||
			!strcasecmp(ks_monitor,  monitorabbrev_name[i])) {
			sel = i;
			break;
		    }

		if (sel != -1) {
		    state->hsyncrange = monitortype_range[sel];
		    if (sel < 4)
			/* use 60 hz for low handwidth monitors */
			state->vsyncrange = "50-61";
		    else if (sel < 7)
			state->vsyncrange = "50-70";
		    else
			state->vsyncrange = "50-90";
		    state->dpms = 1;
		    state->monitoridentifier = monitortype_name[sel];
		    state->monitorvendorname = "Unknown";
		    state->monitormodelname = "Unknown";
		    return 0;
		}
	    }
	} else {
	    /* maybe they gave horizontal and vertical sync rates */
	    if (ks_hsync && ks_vsync) {
		state->hsyncrange = ks_hsync;
		state->vsyncrange = ks_vsync;
		state->dpms = 1;
		state->monitoridentifier = "User Specified Specs";
		state->monitorvendorname = "Unknown";
		state->monitormodelname = "Unknown";
		return 0;
	    }
	}


	if (sel == -1 || !ks_monitor) {
	    /* Assume standard VGA monitor. */
	    state->hsyncrange = "31.5" ;
	    state->vsyncrange = "60";
	    state->dpms = 1;
	    state->monitoridentifier = "Standard VGA, 640x480 @ 60 Hz";
	    state->monitorvendorname = "Unknown";
	    state->monitormodelname = "Unknown";
	    return 0;
	}
    } /* kickstart handler */

    /* Choose from available monitors in the database */
    monitors = alloca(sizeof(*monitors) * (monitorInfoNum + 2));
    monitors[0] = _("Custom");
    for (i = 0, j = 0; monitorInfo[i].name; i++)
      if (strcmp(monitors[j], monitorInfo[i].name))
	monitors[++j] = monitorInfo[i].name;
    monitors[++j] = NULL;

    i = j = 0;
    newtGetScreenSize(&cols, &rows);
    if (rows <= 25)
	rows = 7;
    else
	rows -= 18;
    rc = newtWinMenu(_("Monitor Setup"),
		   _("What type of monitor do you have? If you would rather "
		   "specify the sync frequencies of your monitor, choose "
		   "\"Custom\" from the list."),
		     56, 0, 20, rows, monitors, &j, _("Ok"), _("Back"), NULL);
    if (rc == 2) return 1;

    if (!j) {
	return custommonitor_configuration(state, 1);
    } else {
      for (i = 0; monitorInfo[i].name; i++)
	if (!strcmp(monitorInfo[i].name, monitors[j]))
	  break;
	state->hsyncrange = monitorInfo[i].horizSync;
	state->vsyncrange = monitorInfo[i].vertSync;
	state->dpms = monitorInfo[i].dpms;
	state->monitoridentifier = monitorInfo[i].name;
	state->monitorvendorname = "Unknown";
	state->monitormodelname = "Unknown";
    }
    return 0;
}

int customsyncrate_configuration(struct XconfigState *state) {

    newtComponent form, ok, back, answer, textbox;
    newtComponent hsync_entry, vsync_entry;
    newtGrid grid, buttons, subgrid;
    char *hsync_val, *vsync_val;

    textbox = newtTextboxReflowed(-1, -1, 
				  _(" Please enter the horizontal and "
				    "vertical sync ranges for your monitor. "
				    "This information should be available "
				    "from your monitor manual."), 
				  50, 10, 20, 0);
    form = newtForm(NULL, NULL, 0);

    subgrid = newtCreateGrid(3, 2);

    hsync_entry = newtEntry(-1, -1, state->hsyncrange, 10, &hsync_val, 0);
    vsync_entry = newtEntry(-1, -1, state->vsyncrange, 10, &vsync_val, 0);

    newtGridSetField(subgrid, 0, 0, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("Horizontal Sync:")),
		     0, 0, 2, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridSetField(subgrid, 0, 1, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("Vertical Sync:")),
		     0, 0, 2, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridSetField(subgrid, 1, 0, NEWT_GRID_COMPONENT,
    		     hsync_entry,
		     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridSetField(subgrid, 1, 1, NEWT_GRID_COMPONENT,
    		     vsync_entry,
		     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridSetField(subgrid, 2, 0, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("kHz")),
		     1, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridSetField(subgrid, 2, 1, NEWT_GRID_COMPONENT,
    		     newtLabel(-1, -1, _("Hz")),
		     1, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

    newtGridAddComponentsToForm(subgrid, form, 1);

    buttons = newtButtonBar(_("OK"), &ok, _("Back"), &back, NULL);

    grid = newtGridBasicWindow (textbox, subgrid, buttons);
    newtGridWrappedWindow(grid, _("Custom Monitor Information"));

    newtFormAddComponent(form, textbox);

    newtGridAddComponentsToForm(buttons, form, 1);
    newtGridFree(grid, 1);

    answer = newtRunForm(form);

    newtPopWindow();

    if (answer == back){
        newtFormDestroy(form);
	return 1;
    }
    else {
	state->hsyncrange = strdup(newtEntryGetValue(hsync_entry));
	state->vsyncrange = strdup(newtEntryGetValue(vsync_entry));
	state->dpms = 1;
	state->monitoridentifier = "CustomConfiguration";
	state->monitorvendorname = "Unknown";
	state->monitormodelname = "Unknown";
        newtFormDestroy(form);
	return 0;
    }
}


/* return 1 on cancel, 2 on error, 0 on success */
int custommonitor_configuration(struct XconfigState *state, int dir) {
    int rc;
    static int hsync = 0;
    static int vsync = 0;
    static int customrange = 0;
    static char hsync_val[255] = "\0";
    static char vsync_val[255] = "\0";

    enum { START, HSYNC, VSYNC, CUSTOMRATES, MONITORDONE } step = START;

    if (dir == -1) {
	if (customrange)
	    step = CUSTOMRATES;
	else
	    step = VSYNC;
    }
    
    while (step != MONITORDONE) {
      switch (step) {
      case START:
	rc = newtWinChoice(_("Custom Monitor Setup"), _("Ok"), _("Back"),
			   _(monitorintro_text));
	if (rc == 2)
	  return 1;
	else
	  step++;
	break;

      case HSYNC:
	rc = newtWinMenu(_("Custom Monitor Setup (Continued)"),
			 _(hsyncintro_text),
			 60, 0, 15, 4, monitortype_name, &hsync,
			 _("Ok"), _("Back"), NULL);
	if (rc == 2)
	  step--;
	else {
	    if (hsync) {
		state->hsyncrange = monitortype_range[hsync];
		step++;
		customrange = 0;
	    } else {
		step = CUSTOMRATES;
		customrange = 1;
	    }
	}
	break;
      case VSYNC:
	rc = newtWinMenu(_("Custom Monitor Setup (Continued)"),
			 _(vsyncintro_text),
			 60, 0, 15, 4, vsync_range, &vsync,
			 _("Ok"), _("Back"), NULL);
	if (rc == 2)
	  step--;
	else {
	  state->vsyncrange = strdup(vsync_range[vsync]);
	  *(strchr(state->vsyncrange, ' ')) = '\0';

	  state->monitoridentifier = "My Monitor";
	  state->monitorvendorname = "Unknown";
	  state->monitormodelname = "Unknown";
	  state->dpms = 1;
	  step = MONITORDONE;
	  custom = 1;
	  break;
	}
      case CUSTOMRATES:
	  state->hsyncrange = strdup(hsync_val);
	  state->vsyncrange = strdup(vsync_val);
	  rc = customsyncrate_configuration( state );

	  if (rc) {
	      step = HSYNC;
	  } else {
	      strncpy(hsync_val, state->hsyncrange, sizeof(hsync_val));
	      strncpy(vsync_val, state->vsyncrange, sizeof(vsync_val));
	      step = MONITORDONE;
	      custom = 1;
	  }

	  break;

      case MONITORDONE:
	custom = 1;
	break;
      }
    }
    return 0;
}

/*
 * Card database.
 */

#define NU_ACCELSERVER_IDS 11

static char *accelserver_id[NU_ACCELSERVER_IDS] = {
	"S3",
	"Mach32",
	"Mach8",
	"8514",
	"P9000",
	"AGX",
	"W32",
	"Mach64",
	"I128",
	"S3V",
	"3DLabs"
};

void get_out() {
	newtFinished();
	exit(2);
}

void ListCardTypes(void) {
    int i;
    if (parse_database()) {
	fprintf(stderr, _("\n\nCan't read card database.  \n\n"));
	return;
    }
    for (i=0; i < lastcard; i++)
	fprintf(stdout,"                        %s\n",card[i].name);
}

void ListServerTypes(void) {
    int i;
#ifdef __sparc__
    for (i=0; i < NU_SUNSERVERS; i++)
	fprintf(stdout,"                            %s\n",sunservers[i]);
    if (check_sunv9())
#endif
        for (i=0; i < NU_ALLSERVERS; i++)
	    fprintf(stdout,"                            %s\n",allservers[i]);
}

int carddb_configuration(struct XconfigState *state, char *read_card) {
    int i, j, k, n = 0;
    int user_card_selected, user_chose_card;
    newtComponent form, ok, answer, textbox;
    newtGrid grid, buttons, subgrid;
    char buf[80];
    char *server_name,*driver_name;
    char *driver;
    char ** cards;
    struct device **devs;
    int rc;
    int serverok = 0;

    driver_name=NULL;

    /* Fixup for the old way to specify servers */
    if (read_card) {
	/* we know that the Xorg server is good... */
	if (!strcmp (read_card, "Xorg"))
	    serverok = 1;
	else {
	    /* check to see if its listed in the servers */
	    for (i = 0; i < NU_ALLSERVERS; i++){
		if (!strcasecmp(read_card, allservers[i])) {
		    serverok = 1;
		    break;
		}
	    }
#ifdef __sparc__
	    for (i = 0; i < NU_SUNSERVERS; i++){
		if (!strcasecmp(read_card, sunservers[i])) {
		    serverok = 1;
		    break;
		}
	    }
#endif
	}
	
	if (!serverok) {
	    /* check to see if what was passed matches
	       the servers we know about with the XF86_ lopped
	       off the front.  This is the way we used to configure
	       the server from kickstart... */
	    for (i = 0; i < NU_ALLSERVERS; i++) {
		if (!strcasecmp(read_card, allservers[i] + 5)) {
		    read_card = allservers[i];
		    serverok = 1;
		    break;
		}
	    }
	}
#ifdef __sparc__
	    /* check to see if what was passed matches
	       the sun servers we know about with the Xs lopped
	       off the front.  This is the way we used to configure
	       the server from kickstart... */
	if (!serverok) {
	    for (i = 0; i < NU_SUNSERVERS; i++) {
		if (!strcasecmp(read_card, sunservers[i] + 1)) {
		    read_card = sunservers[i];
		    serverok = 1;
		    break;
		}
	    }
	}
#endif
    }
    
    
    /* are we continuing an --pick with a --continue */
    /* card_selected is already set                  */
    if (rhcontinue) {
        if(strcmp("-1",read_card)==0){
            read_card=NULL;
            goto probe_rhcontinue;
        }
        
	if (card_selected != UNLISTED_CARD){
            if(card[card_selected].server) {
		server_name = malloc (strlen (card[card_selected].server) + 6);
		strcpy (server_name, "XF86_");
		strcat (server_name, card[card_selected].server);
	    }
            if(card[card_selected].driver!=NULL)
                driver_name = strdup(card[card_selected].driver);
	    //printf("driver_name: %s\n", driver_name);
        }
	else
	    server_name = strdup(read_card);
	
	 if (driver_name) {
            if(driver_name){
                state->xf4driver=strdup(driver_name);
                free(driver_name);
            }
	    state->screentype = 4;
	} else if (server_name) {
	 
	    if (strcmp(server_name, "XF86_Mono") == 0)
		card_screentype = 1;
	    else if (strcmp(server_name, "XF86_VGA16") == 0)
		card_screentype = 2;
	    else if (strcmp(server_name, "XF86_SVGA") == 0)
		card_screentype = 3;
#ifdef __sparc__
	    else if (strncmp(server_name, "xsun", 4) == 0) {
	        card_screentype = 5;
	        for (i = 0; i < NU_SUNSERVERS; i++)
	            if (strcmp(server_name, sunservers[i]) == 0) {
	                card_accelserver = i;
	                break;
	            }
	    }
#endif
	    else
		for (i = 0; i < NU_ACCELSERVER_IDS; i++)
		    if (strcmp(server_name, accelserver_id[i]) == 0) {
			card_screentype = 4;
			card_accelserver = i;
			break;
		    }

	    strcpy(state->serverbuf,server_name);
            if(driver_name){
                state->xf4driver=strdup(driver_name);
                free(driver_name);
            }
	    state->screentype = card_screentype;
	} else {
	    newtWinMessage(_("Ok"), _("Unknown Server"),
		_("Somehow in continue mode an unknown server type exists."));
	    newtFinished();
	    exit(2);
	}

	return 0;
    }


    /* did they specify kickstart and give us a card or server type to */
    /* use?                                                            */
    if (ks_card && !read_card) {
	/* find a match */
	for (i = 0; i < lastcard; i++)
	    if (!strcasecmp(ks_card, card[i].name)) {
		read_card = card[i].server;
                if(card[i].driver)
                    driver_name = strdup(card[i].driver);
		card_selected = i;
		break;
	    }
    }

    /* maybe they gave a server type instead? */
    if ((ks_server || ks_driver) && !read_card) {
	/* find a match */
#ifdef __sparc__
	for (i = 0; i < NU_SUNSERVERS; i++)
            if(ks_server){
                if (!strcasecmp(ks_server, sunservers[i])) {
                    read_card = sunservers[i];
                    card_selected = UNLISTED_CARD;
                    break;
                }
            }
	if (i == NU_SUNSERVERS && check_sunv9())
#endif
	    if(ks_server){
		for (i = 0; i < NU_ALLSERVERS; i++){
		    if (!strcasecmp(ks_server, allservers[i])) {
			read_card = allservers[i];
			card_selected = UNLISTED_CARD;
			break;
		    }
		}
            }
    }

    /* hmm, they havent given use any help  */
    /* msf - ugly but easiest way to go now */
    /* if no card specified, try pci probe  */

  probe_rhcontinue:
    if (!read_card) {
	int i;

	initializeBusDeviceList(BUS_UNSPEC);
	devs = probeDevices(CLASS_VIDEO,BUS_UNSPEC,0);
        if (devs) {
	    char *val=NULL;
	    char buf[64];
	    int last_server = 0;
		
	    for (j=0; devs[j] != NULL; j++)
		  if (devs[j]->bus == BUS_SBUS) {
			for (k=0; k < lastcard; k++) {
                        	if (!strcmp(devs[j]->desc, card[k].name)) {
					driver_name = strdup(card[k].driver);
				}	 
                        } 
			  if (!strncmp(devs[j]->driver, "Server:", 7)) {
			          val = alloca(6 + strlen (devs[j]->driver+7));
				  sprintf (val, "%s", devs[j]->driver+7);
				  /* printf(val, "%s", devs[j]->driver+7); */

				  /* find a match */
				  for (i = 0; i < NU_SUNSERVERS; i++)
				    if (!strcmp(val, sunservers[i])) {
					    if (!read_card || last_server < i) {
						    read_card = sunservers[i];
						    last_server = i;
						    card_selected = UNLISTED_CARD;
						    pciprobedcard = strdup(devs[j]->desc);
					    }
					    n=j;
					    break;
				    }
			  }
		  } else if (devs[j]->bus == BUS_PCI) { 
			  if (strcasecmp(devs[j]->driver, "UNKNOWN")) {
				  /* did we find a server or a card type */
				  if (!strncmp(devs[j]->driver,"Card:",5)) {
					  val = devs[j]->driver+5;

					  /* find a match */
					  for (i = 0; i < lastcard; i++)
                                              
                                              if (!strcmp(val, card[i].name)) {
                                                  if(card[i].server){
                                                      read_card = alloca (strlen (card[i].server) + 6);
                                                      strcpy (read_card, "XF86_");
                                                      strcat (read_card, card[i].server);
                                                  }
                                                  if(card[i].driver){
                                                      driver_name = strdup(card[i].driver);
                                                      if(!(card[i].server))
                                                         read_card=strdup("None");
                                                      if((card[i].flags & UNSUPPORTED) && preferxf3 ){
                                                          preferxf3=1;
                                                      }
                                                  }
                                                  card_selected = i;
                                                  pciprobedcard = strdup(devs[j]->desc);
                                                  state->pciid=((struct pciDevice*)devs[j])->deviceId;
                                                  state->pcivendorid=((struct pciDevice*)devs[j])->vendorId;
                                                  n=j;
                                                  break;
                                              }
				  } else if (!strncmp(devs[j]->driver, "Server:", 7)) {
				          val = alloca(6 + strlen (devs[j]->driver+7));
					  sprintf (val, "XF86_%s", devs[j]->driver+7);

					  /* find a match */
					  for (i = 0; i < NU_ALLSERVERS; i++)
					    if (!strcmp(val, allservers[i])) {
						    read_card = allservers[i];
						    card_selected = UNLISTED_CARD;
						    pciprobedcard = strdup(devs[j]->desc);
						    n=j;
						    break;
					    }
				  }
			  }
		  }
		
	    buf[0] = 0;
	    if (devs[n] && devs[n]->bus == BUS_SBUS) {
		    struct sbusDevice *s = (struct sbusDevice *)devs[n];
		    char *p;

		    sprintf(buf, _("  Resolution: %dx%d"), s->width, s->height);
		    p = strchr(buf, 0);
		    if (s->freq)
		      sprintf(p, "x%d", s->freq);
		    p = strchr(p, 0);
		    if (s->monitor != 0)
		      sprintf(p, _("  Monitor type %d\n"), s->monitor);
		    else
		      sprintf(p, "\n");
	    }

	    if ((!kickstart && read_card)&&(!rhcontinue)) {
		    buttons = newtButtonBar(_("Ok"), &ok, NULL);
		    textbox = newtTextboxReflowed(-1, -1, 
 (devs[n]->bus == BUS_SBUS) ? _("SBUS probing found:") : _("PCI probing found:"), 50, 10, 20, 0);

		    form = newtForm(NULL, NULL, 0);

		    subgrid = newtCreateGrid(3, 3);
		    
		    newtGridSetField(subgrid, 0, 0, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1, 
	      (devs[n]->bus == BUS_SBUS) ? _("SBUS Entry   :") : _("PCI Entry    :")),
				     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
/*		    newtGridSetField(subgrid, 0, 1, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1, _("X Server     :")),
				     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0); */
		    newtGridSetField(subgrid, 0, 2, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1, _("Xorg driver:")),
				     0, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);
                    if(pciprobedcard && (strlen(pciprobedcard)>40)){
                        pciprobedcard[40]=0;
                    }
                    newtGridSetField(subgrid, 1, 0, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1, 
					       (read_card) ? pciprobedcard : _("No match")),
				     1, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

/*		    newtGridSetField(subgrid, 1, 1, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1,(read_card) ? read_card : _("No match")),
				     1, 0, 0, 0, NEWT_ANCHOR_LEFT, 0); */
		    if (card_selected != UNLISTED_CARD && card[card_selected].flags & UNSUPPORTED){
                        driver = alloca (strlen(driver_name) + strlen(_(" (Not used by default)")) + 1);
                        sprintf (driver, "%s%s", driver_name, _(" (Not used by default)"));
                        if((preferxf3) && (card[card_selected].flags & UNSUPPORTED)){
                            preferxf3=1;
                        }
		    } else
			driver = driver_name;
		    newtGridSetField(subgrid, 1, 2, NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1,(driver_name) ? (driver) : _("No match")),
				     1, 0, 0, 0, NEWT_ANCHOR_LEFT, 0);

		    newtGridSetField(subgrid, (preferxf3 ? 1 : 2), 2,
				     NEWT_GRID_COMPONENT,
				     newtLabel(-1, -1, _("(default)")),
				     1, 0, 0, 0,
				     NEWT_ANCHOR_LEFT, 0);
                    
		    newtGridAddComponentsToForm(subgrid, form, 1);

		    grid = newtGridBasicWindow(textbox, subgrid, buttons);
		    newtGridWrappedWindow(grid, (devs[n]->bus == BUS_SBUS) ? _("SBUS Probe") : _("PCI Probe"));
		    newtFormAddComponent(form, textbox);
		    newtGridAddComponentsToForm(buttons, form, 1);
		    newtGridFree(grid, 1);

		    answer = newtRunForm(form);
		    newtPopWindow();
		    newtFormDestroy(form);
	    }
	    for (i=0; devs[i] ; i++)
		  devs[i]->freeDevice(devs[i]);
	    free(devs);
		
	}
    }

    /* if we got here and found nothing, and we are doing a kickstart */
    /* we just give up and use VGA16 server                           */
    if (kickstart && !read_card) {
	card_selected = UNLISTED_CARD;
	read_card ="VGA16";
    }

    user_card_selected = -1;
    user_chose_card = 0;
    if (!read_card || (expert && read_card && card_selected!=UNLISTED_CARD)) {
	cards = alloca(sizeof(*cards) * (lastcard + 2));
	for (i = 0; i < lastcard; i++) {
	    snprintf(buf, 70, "%-49s%s", card[i].name, card[i].chipset);
	    cards[i] = strdup(buf);
	}
	cards[i] = _("Unlisted Card");
	cards[i + 1] = NULL;

	if (pciprobedcard)
	    user_card_selected = card_selected;

	rc = newtWinMenu(_("Choose a Card"),
			    _("Pick a Card from the list below "
			      "(Or choose \"Unlisted Card\" at the bottom of "
			      "the list if your card isn't listed):"),
			 45, 0, 15, 9, cards, &user_card_selected,
			_("Ok"), _("Back"), NULL);

	if (rc == 2) return 1;
	if (user_card_selected == i)
	    user_card_selected = UNLISTED_CARD;
	user_chose_card = 1;

	for (i = 0; i < lastcard; i++)
	    free(cards[i]);
    }

    /*
     * Look at the selected card.
     */
    server_name = NULL;
    if (user_chose_card) {
	if (user_card_selected != UNLISTED_CARD) {
            if(card[user_card_selected].server)
                server_name = strdup(card[user_card_selected].server);
            if(card[user_card_selected].driver)
                driver_name = strdup(card[user_card_selected].driver);
	    card_selected = user_card_selected;
	} else {
	    card_selected = UNLISTED_CARD;
	}
    } else if (pciprobedcard) {
	if (card_selected != UNLISTED_CARD){
            if(card[card_selected].server)
                server_name = strdup(card[card_selected].server);
            if(card[card_selected].driver)
                driver_name = strdup(card[card_selected].driver);
        }
	else if (read_card)
	    server_name = strdup(read_card);
    } else if (read_card)
	    server_name = strdup(read_card);

    if (server_name) {
	if (strcmp(server_name, "Mono") == 0)
	    card_screentype = 1;
	else if (strcmp(server_name, "VGA16") == 0)
		card_screentype = 2;
	else if (strcmp(server_name, "SVGA") == 0)
		card_screentype = 3;
#ifdef __sparc__
	else if (strncmp(server_name, "Sun", 3) == 0) {
	    card_screentype = 5;
	    for (i = 0; i < NU_SUNSERVERS; i++)
	        if (strcmp(server_name, sunservers[i]) == 0) {
	            card_accelserver = i;
	            break;
	        }
	}
#endif
	else
	    for (i = 0; i < NU_ACCELSERVER_IDS; i++)
		if (strcmp(server_name, accelserver_id[i]) == 0) {
		    card_screentype = 4;
		    card_accelserver = i;
		    break;
		}
    }

    if ((card_selected == UNLISTED_CARD) &&
	(!read_card || (expert && read_card))) {
	state->servernum = 0;
	rc = newtWinMenu(_("Pick a Server"), _("Which server do you need?"),
			 20, 0, 30, 8, allservers, &state->servernum, _("Ok"),
			 _("Back"), NULL);
	if (rc == 2) get_out();

	strcpy(state->serverbuf, allservers[state->servernum]);
	if (state->servernum < 3)
	    card_screentype = state->servernum + 1;
	else
	    card_screentype = 4;
    } else {
	if (!read_card){
            if(card[card_selected].server) {
/*  		printf ("xxxxxxxxxxxxxxxxxxxxxxxx %s\n", card[card_selected].server); */
                strcpy(state->serverbuf, card[card_selected].server);
	    }
            if(card[card_selected].driver)
                state->xf4driver=strdup(card[card_selected].driver);
            if(card[card_selected].name)
                state->deviceidentifier=strdup(card[card_selected].name);
        }
	else{
	    strcpy(state->serverbuf, read_card);
            if(driver_name){
                state->xf4driver=strdup(driver_name);
                free(driver_name);
            }
            if(card[card_selected].name)
                state->deviceidentifier=strdup(card[card_selected].name);
        }
    }
    state->screentype = card_screentype;

    return 0;
}
    


/*
 * Screen/video card configuration.
 */


static char *ramdac_name[] = {
	N_("No RAMDAC Setting (recommended)"),
	N_("AT&T 20C490 (S3 and AGX servers, ARK driver)"),
	N_("AT&T 20C498/21C498/22C498 (S3, autodetected)"),
	N_("AT&T 20C409/20C499 (S3, autodetected)"),
	N_("AT&T 20C505 (S3)"),
	N_("BrookTree BT481 (AGX)"),
	N_("BrookTree BT482 (AGX)"),
	N_("BrookTree BT485/9485 (S3)"),
	N_("Sierra SC15025 (S3, AGX)"),
#if XFREE86_VERSION >= 311
	N_("S3 GenDAC (86C708) (autodetected)"),
	N_("S3 SDAC (86C716) (autodetected)"),
#else
	N_("S3 GenDAC (86C708)"),
	N_("S3 SDAC (86C716)"),
#endif
	N_("STG-1700 (S3, autodetected)"),
	N_("STG-1703 (S3, autodetected)"),
	N_("TI 3020 (S3)"),
	N_("TI 3025 (S3, autodetected)"),
	N_("TI 3026 (S3, autodetected)"),
        N_("IBM RGB 514 (S3, autodetected)"),
        N_("IBM RGB 524 (S3, autodetected)"),
        N_("IBM RGB 525 (S3, autodetected)"),
        N_("IBM RGB 526 (S3)"),
        N_("IBM RGB 528 (S3, autodetected)"),
        N_("ICS5342 (S3, ARK)"),
        N_("ICS5341 (W32)"),
        N_("IC Works w30C516 ZoomDac (ARK)"),
	N_("Normal DAC"),
	NULL
};
#define NU_RAMDACS ((sizeof(ramdac_name) / sizeof(*ramdac_name)) - 1)

static char *ramdac_id[NU_RAMDACS] = {
	NULL,
	"att20c490", "att20c498", "att20c409", "att20c505", "bt481", "bt482",
	"bt485", "sc15025", "s3gendac", "s3_sdac", "stg1700","stg1703",
	"ti3020", "ti3025", "ti3026", "ibm_rgb514", "ibm_rgb524",
        "ibm_rgb525", "ibm_rgb526", "ibm_rgb528", "ics5342", "ics5341",
        "zoomdac", "normal"
};

static char *clockchip_name[] = {
	N_("No Clockchip Setting (recommended)"),
        N_("Chrontel 8391"),
        N_("ICD2061A and compatibles (ICS9161A, DCS2824)"),
        N_("ICS2595"),
        N_("ICS5342 (similar to SDAC, but not completely compatible)"),
        N_("ICS5341"),
        N_("S3 GenDAC (86C708) and ICS5300 (autodetected)"),
        N_("S3 SDAC (86C716)"),
        N_("STG 1703 (autodetected)"),
        N_("Sierra SC11412"),
        N_("TI 3025 (autodetected)"),
        N_("TI 3026 (autodetected)"),
        N_("IBM RGB 51x/52x (autodetected)"), NULL
};
#define NU_CLOCKCHIPS ((sizeof(clockchip_name) / sizeof(*clockchip_name)) - 1)

static char *clockchip_id[NU_CLOCKCHIPS] = {
        NULL, "ch8391", "icd2061a", "ics2595", "ics5342", "ics5341",
        "s3gendac", "s3_sdac",
        "stg1703", "sc11412", "ti3025", "ti3026", "ibm_rgb5xx",
};

static char *probeonlywarning_text =
N_("It is possible that the hardware detection routines in the server somehow "
"cause the system to crash and the screen to remain blank. If this is the "
"case, skip this step the next time. The server may need a "
"Ramdac, ClockChip or special option (e.g. \"nolinear\" for S3) to probe "
"and start-up correctly.");


static int videomemory[] = {
    256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072
};
#define NUM_VIDEOMEMORY (sizeof(videomemory) / sizeof(*videomemory))

static char *wangermemorys[NUM_VIDEOMEMORY + 1] = {
    "256 KB", "512 KB", "1 MB", "2 MB", "4 MB",
    "8 MB", "16 MB", "32 MB", "64 MB", "128 MB", NULL
};

/* On entry, "modes" contains a *single* mode, which is the highest
   resolution we can do for this bpp. We return a list of all of the
   resolutions we can do given our amount of memory, which "modes" as
   the upper limit. */

void memory_prune_modes(struct XconfigState *state, char ** modes, int bpp) {
    int bytesPerPixel = bpp / 8;
    char * possible = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    int maxx, maxy;
    int i, xres, yres;
    char buf[100] = "";
    char ** list = splitString(possible, strlen(possible), ' ');

    modeline_to_res(list[0], &maxx, &maxy);

    for (i = 0; list[i]; i++) {
	modeline_to_res(list[i], &xres, &yres);
	if (((xres * yres * bytesPerPixel) < (state->videomemory * 1024))
		&& ((xres <= maxx) && (yres <= maxy))) {
	    if (buf[0])
		strcat(buf, " ");
	strcat(buf, list[i]);
	}
    }

    if (!strlen(buf))
	*modes = NULL;
    else
	*modes = strdup(buf);

    free(list);
}

int find_legal_modes(struct XconfigState *state) {
    /* Walk through the config modelines and remove those which we */
    /* don't have enough memory for */
  /*
   * Initialize screen mode variables for svga and accel
   * to default values.
   * XXXX Doesn't leave room for off-screen caching in 16/32bpp modes
   *      for the accelerated servers in some situations.
   */
  state->modesline8bpp =
    state->modesline16bpp =
    state->modesline32bpp = "\"640x480\"";
  if (state->videomemory >= 8192) {
    state->modesline8bpp = state->modesline16bpp = state->modesline32bpp =  "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
  } else if (state->videomemory >= 4096) {
    state->modesline8bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline32bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
  } else if (state->videomemory >= 2048) {
    state->modesline8bpp = "\"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline32bpp = "\"800x600\" \"640x480\"";
  } else if (state->videomemory >= 1024) {
    state->modesline8bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"800x600\" \"640x480\"";
    state->modesline32bpp = "\"640x400\"";
  } else if (state->videomemory >= 512) {
    state->modesline8bpp = "\"800x600\" \"640x480\"";
    state->modesline16bpp = "\"640x400\"";
  } else if (state->videomemory >= 256) {
    state->modesline8bpp = "\"640x480\"";
  }
 
  if (state->videomemory < 256) {
    if (!kickstart) {
      newtWinMessage(_("X Error"), _("Ok"),
		     _("There was an error detecting "
		       "the video ram on your card. You should try "
		       "configuring the video card manually."));
      get_out();
    } else {
      fprintf(stderr, _("Fatal error: Invalid amount of video "
			"memory.\n"));
      exit(2);
    }
  }
  
  memory_prune_modes(state, &state->modesline8bpp, 8);
  memory_prune_modes(state, &state->modesline16bpp, 16);
  memory_prune_modes(state, &state->modesline32bpp, 32);
  return 0;
}

int find_all_legal_modes(struct XconfigState *state) {
  /*
   * Initialize screen mode variables for svga and accel
   * to default values.
   * XXXX Doesn't leave room for off-screen caching in 16/32bpp modes
   *      for the accelerated servers in some situations.
   */
  state->modesline8bpp =
    state->modesline16bpp =
    state->modesline32bpp = "\"640x480\"";
  if (state->videomemory >= 8192) {
    state->modesline8bpp = state->modesline16bpp = state->modesline32bpp =  "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
  } else if (state->videomemory >= 4096) {
    state->modesline8bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline32bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
  } else if (state->videomemory >= 2048) {
    state->modesline8bpp = "\"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline32bpp = "\"800x600\" \"640x480\"";
  } else if (state->videomemory >= 1024) {
    state->modesline8bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
    state->modesline16bpp = "\"800x600\" \"640x480\"";
    state->modesline32bpp = "\"640x400\"";
  } else if (state->videomemory >= 512) {
    state->modesline8bpp = "\"800x600\" \"640x480\"";
    state->modesline16bpp = "\"640x400\"";
  } else if (state->videomemory >= 256) {
    state->modesline8bpp = "\"640x480\"";
  } else {
    if (!kickstart) {
      newtWinMessage(_("X Error"), _("Ok"),
		     _("There was an error detecting "
		       "the video ram on your card. You should try "
		       "configuring the video card manually."));
      return 1;
    } else {
      fprintf(stderr, _("Fatal error: Invalid amount of video "
			"memory.\n"));
      exit(-1); /* what can we do in kickstart? No 'ASK' option yet */
    }
  }
  return 0;
}


int screen_configuration(struct XconfigState *state) {
    int i, vidMemNum;
    int skipramdac,skipclock = 0;
    newtComponent form, ok, cancel, answer, textbox;
    newtGrid grid, buttons;
    char tfilenm1[]="/tmp/Xprobe1-XXXXXX";
    char tfilenm2[]="/tmp/Xprobe2-XXXXXX";
    char tcfgnm[]="/tmp/xf86config-XXXXXX";
    char cmd[250];
    int  tfilefd;
    int  rc;
    GSList *specials;
    enum { MEMORY, RAMDAC, CLOCKCHIP,
	   SKIPCLOCKPROBING, SCREENDONE } step = MEMORY;

    /*
     * Configure the "Device" section for the video card.
     */
    while (step != SCREENDONE) {
      switch (step) {

      case MEMORY:
          vidMemNum = 2;
          if (probedmemory > 0) {
              for(i=0; i < NUM_VIDEOMEMORY;i++){
                  if (videomemory[i] <= probedmemory) {
                      vidMemNum=i;
                  }
              }
          }

          /* Special case for i810, since it doesn't have real video memory */
          
          if( (videomemory[vidMemNum]<16384)
	      && state->xf4driver
	      && (strcasestr(state->xf4driver,"i810")!=NULL) ){
              vidMemNum=6;
          }
              
          rc = newtWinMenu(_("Video Memory"),
                           _("How much video memory do you have?"),
                           30, 0, 30, 8, wangermemorys, &vidMemNum, _("Ok"),
                           _("Back"), NULL);
          if (rc == 2) {
              return 1;
          }
          state->videomemory = videomemory[vidMemNum];

          if (pciprobedcard && !state->deviceidentifier)
              state->deviceidentifier = pciprobedcard;
          else if(!state->deviceidentifier)
              state->deviceidentifier = "My Video Card";
          state->devicevendorname = "Unknown";
          state->deviceboardname = "Unknown";
          
        /* Check for special combinations... ugly ugly ugly*/
        
        specials=loadComboFile(MONITORCARDCOMBO_DATABASE_FILE);
        if(specials!=NULL){
            combo=findCombo(specials,state);
            if(combo!=NULL){
                state->modesline8bpp=get8bitModes(combo);
                state->modesline16bpp=get16bitModes(combo);
                state->modesline32bpp=get32bitModes(combo);
            }
        }


        if(combo==NULL){

	  /*
	   * Initialize screen mode variables for svga and accel
	   * to default values.
	   * XXXX Doesn't leave room for off-screen caching in 16/32bpp modes
	   *      for the accelerated servers in some situations.
	   */
	  state->modesline8bpp =
	    state->modesline16bpp =
	    state->modesline32bpp = "\"640x480\"";
	  if (state->videomemory >= 8192) {
	    state->modesline8bpp = state->modesline16bpp = state->modesline32bpp =  "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	  } else if (state->videomemory >= 4096) {
	    state->modesline8bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	    state->modesline16bpp = "\"1600x1200\" \"1400x1050\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	    state->modesline32bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	  } else if (state->videomemory >= 2048) {
	    state->modesline8bpp = "\"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	    state->modesline16bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	    state->modesline32bpp = "\"800x600\" \"640x480\"";
	  } else if (state->videomemory >= 1024) {
	    state->modesline8bpp = "\"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
	    state->modesline16bpp = "\"800x600\" \"640x480\"";
	    state->modesline32bpp = "\"640x400\"";
	  } else if (state->videomemory >= 512) {
	    state->modesline8bpp = "\"800x600\" \"640x480\"";
	    state->modesline16bpp = "\"640x400\"";
	  } else if (state->videomemory >= 256) {
	    state->modesline8bpp = "\"640x480\"";
	  } else {
	    printf(_("Fatal error: Invalid amount of video memory.\n"));
	    exit(-1);
	  }
	  
        }
        
        
	/*
	 * Handle the Ramdac/Clockchip setting.
	 */

	skipramdac = 0;

	if (state->servernum < 3)
	  skipramdac = 1;
	if (card_selected != UNLISTED_CARD)
	  if (card[card_selected].ramdac != NULL) {
	    state->ramdac = card[card_selected].ramdac;
	    skipramdac = 1;
	  }
	step = RAMDAC;
	break;

      case RAMDAC:
	if (skipramdac != 1) {
	  i = 0;
	  rc = newtWinMenu(_("RAMDAC Configuration"),
			   _("Which RAMDAC do you have?"),
			   40, 0, 20, 7, ramdac_name, &i,
			   _("Ok"), _("Back"), NULL);
	  if (rc == 2) get_out();
	  state->ramdac = ramdac_id[i];
	}

	if (card_selected != UNLISTED_CARD) {
	    state->clockchip = card[card_selected].clockchip;
	    if(state->clockchip != NULL)
		skipclock = 1;
	}
	step = CLOCKCHIP;
	break;

      case CLOCKCHIP:
	if (skipclock != 1) {
	  i = 0;
	  rc = newtWinMenu(_("Clockchip Configuration"),
			   _("Which Clockchip do you have?"), 40, 0, 30, 8,
			   clockchip_name, &i, _("Ok"), _("Back"), NULL);
	  if (rc ==2 ) {
	    step = MEMORY;
	    break;
	  }

	  state->clockchip = clockchip_id[i];
	}

	/*
	 * Optionally run X -probeonly to figure out the clocks.
	 */

	state->numberofclockslines = 0;

	if (card_selected != UNLISTED_CARD)
	  if (card[card_selected].flags & NOCLOCKPROBE) {
	    step = SKIPCLOCKPROBING;
	    break;
	  }


	if (state->clockchip != NULL) {
	  step = SKIPCLOCKPROBING;
	  break;
	}

	buttons = newtButtonBar(_("Probe"), &ok, _("Skip"), &cancel, NULL);
	textbox = newtTextboxReflowed(-1, -1, _(probeonlywarning_text), 50,
				      10, 20, 0);
	grid = newtGridSimpleWindow(newtLabel(-1, -1,
			      _("Do you want to run 'X -probeonly' now?")),
				    textbox, buttons);
	newtGridWrappedWindow(grid, _("Probe for Clocks"));
	form = newtForm(NULL, NULL, 0);
	newtGridAddComponentsToForm(grid, form, 1);
	newtGridFree(grid, 1);

	answer = newtRunForm(form);
	newtPopWindow();
	newtFormDestroy(form);

	if (answer != cancel) {
	  /*
	   * Write temporary XF86Config and run X -probeonly.
	   * Only allow when root.
	   */
	  FILE *f;
	  char *buf;
	  if (!(tfilefd=mkstemp(tcfgnm))) {
	    cpfailed(1);
	  }
	  close(tfilefd);
	  write_XF86Config(state, tcfgnm);
#ifndef __EMX__
	  sync();
#endif
	  /* make all the required tmp files and close them */
	  /* we'll have shell append to them as we need them*/
	  if (!(tfilefd=mkstemp(tfilenm1))) {
	    unlink(tcfgnm);
	    cpfailed(1);
	  }
	  close(tfilefd);
	  if (!(tfilefd=mkstemp(tfilenm2))) {
	    unlink(tcfgnm);
	    unlink(tfilenm1);
	    cpfailed(1);
	  }
	  close(tfilefd);
	  snprintf(cmd, sizeof(cmd),
		   "/etc/X11/X :9 -probeonly -verbose -config %s 2>> %s",
		   tcfgnm, tfilenm1);
	  if (system(cmd)) {
	    unlink(tfilenm1);
	    unlink(tfilenm2);
	    unlink(tcfgnm);
	    system("/sbin/setsysfont 2> /dev/null");
	    step = SKIPCLOCKPROBING;
	    break;
	  }
	  system("/sbin/setsysfont 2> /dev/null");

	  /* Look for 'clocks:' (case sensitive). */
	  snprintf(cmd, sizeof(cmd), "grep clocks\\: %s >> %s",
		   tfilenm1, tfilenm2);
	  if (system(cmd)) {
	    /* cpfailed(2); */
	    unlink(tfilenm1);
	    unlink(tfilenm2);
	    unlink(tcfgnm);

 	    newtWinMessage(_("Clock Probe Failed"), _("Ok"),
			   _("Configuration will proceed using "
			     "default values."));
	    step = SKIPCLOCKPROBING;
	    break;
	  }
	  f = fopen(tfilenm2, "r");
	  buf = malloc(8192);
	  /* Parse lines. */
	  while (fgets(buf, 8192, f) != NULL) {
	    char *clks;
	    clks = strstr(buf, "clocks: ") + 8;
	    if (clks >= buf + 3 && strcmp(clks - 11, "num") == 0)
	      /* Reject lines with 'numclocks:'. */
	      continue;
	    if (clks >= buf + 8 && strcpy(clks - 14, "pixel ") == 0)
	      /* Reject lines with 'pixel clocks:'. */
	      continue;
	    clks[strlen(clks) - 1] = '\0';	/* Remove '\n'. */
	    state->clocksline[state->numberofclockslines] =
	      malloc(strlen(clks) + 1);
	    if(state->clocksline[state->numberofclockslines]==NULL)
	    {
	    	fputs(_("Out of memory\n"), stderr);
	    	exit(1);
	    }
	    strcpy(state->clocksline[state->numberofclockslines],
		   clks);
	    state->numberofclockslines++;
	  }
	  fclose(f);
	  unlink(tfilenm1);
	  unlink(tfilenm2);
	  unlink(tcfgnm);
	  printf("clocks %s %d\n", state->clocksline[0],
		 state->numberofclockslines);
	  if (state->numberofclockslines == 0)
 	    newtWinMessage(_("Clock Probe Failed"), _("Ok"),
			   _("Configuration will proceed using "
			     "default values."));
	  step = SKIPCLOCKPROBING;
	} else {
	  step = SKIPCLOCKPROBING;
	}
	break;

      case SKIPCLOCKPROBING:
	/*
	 * For the mono and vga16 server, no further configuration is
	 * required.
	 */
	if (state->screentype == 1 || state->screentype == 2)
	  return 0;

	/*
	 * Configure the modes order.
	 */


        
	if (chooseXmodes(state, &state->modesline8bpp, &state->modesline16bpp,
			 &state->modesline32bpp) == 1) {
	  if (!skipramdac)
              step = RAMDAC;
	  else{
              if (!skipclock)
                  step = CLOCKCHIP;
              else
                  step = MEMORY;
          }
	}
        else{
            step = SCREENDONE;
        }
	break;
      case SCREENDONE: break;
      }
    }
    return 0;
}

/* mode line to res */
void modeline_to_res(char *m, int *x, int *y) {
    if (!m || (sscanf(m, "\"%dx%d\"", x, y) != 2))
	*x = *y = 0;

    return;
}


/* see what lines were ok'd by probing */
/* is used by autoprobe_bpp()          */
/* assumes filename points to X -probeobly output */
/* 0 if ok, < 0 on errors              */
int MaxValidMode(char *filename, int *maxx, int *maxy) {
    int *ModeX;
    int *ModeY;
    int nummodes=0;
    int i, x, y;
    int xfree4=0;
    
    FILE *f;

    char buf[200];
    char *s, *t;

    f = fopen(filename, "r");
    if (!f)
	return -1;
    
    while(fgets(buf, sizeof(buf),f)) {
        if((s=strstr(buf,"XFree86 Version 4")))
            xfree4=1;

/* does this line have a ': Mode' and 'mode clock =' on it? */
        if(!xfree4){
            if ((s=strstr(buf, ": Mode")) && strstr(buf,"mode clock =")) {
                s += 7;
                t = strchr(s, ':');
                *t = '\0';
                modeline_to_res(s, &x, &y);
                if (nummodes) {
                    ModeX = (int *) realloc(ModeX, (nummodes+1)*sizeof(int));
                    ModeY = (int *) realloc(ModeY, (nummodes+1)*sizeof(int));
                } else {
                    ModeX = (int *) malloc(sizeof(int));
                    ModeY = (int *) malloc(sizeof(int));
                }
                ModeX[nummodes] = x;
                ModeY[nummodes] = y;
                nummodes++;
            } else if ((s=strstr(buf,"Removing mod")) && strstr(buf,"valid mod")) {
                s += 14;
                t = strchr(s+1, '"')+1;
                *t = '\0';
                modeline_to_res(s, &x, &y);
                for (i=0; i < nummodes; i++) {
                    if (ModeX[i] == x && ModeY[i] == y) {
                        ModeX[i] = 0;
                        ModeY[i] = 0;
                    }
                }
            }
        }
        else{
            if ((s=strstr(buf, "Default mode")) && strstr(buf,"MHz")) {
                s+=13;
                t=strchr(s, ':');
                *t = '\0';
                modeline_to_res(s, &x, &y);
                if (nummodes) {
                    ModeX = (int *) realloc(ModeX, (nummodes+1)*sizeof(int));
                    ModeY = (int *) realloc(ModeY, (nummodes+1)*sizeof(int));
                } else {
                    ModeX = (int *) malloc(sizeof(int));
                    ModeY = (int *) malloc(sizeof(int));
                }
                ModeX[nummodes] = x;
                ModeY[nummodes] = y;
                nummodes++;
                
            }
        }
        
    }
    
    /* we're done, see what max res left is */
    x = y = 0;
    for (i=0; i < nummodes; i++) {
	if (ModeX[i] > x && ModeY[i] > y) {
	    x = ModeX[i];
	    y = ModeY[i];
	}
    }

    *maxx = x;
    *maxy = y;

    if (x > 0 && y > 0)
	return 0;
    else
	return -1;
}


/* see if the currently configuration X server supports the requested */
/* bit depth                                                          */
char *autoprobe_bpp(struct XconfigState *state, int bpp) {
    char buf[255];
    char *mode;
    char *rc;
    char tfilenm1[]="/tmp/Xprobe1-XXXXXX";
    char tcfgnm[]="/tmp/xf86config-XXXXXX";
    char *xf4file=NULL;
    int tfilefd;
    int maxx, maxy;

    rc=NULL;
    /* probe again and see which 8bpp modes were valid */
    if (!(tfilefd=mkstemp(tcfgnm)))
	return NULL;
    close(tfilefd);
    if (!(tfilefd=mkstemp(tfilenm1))) {
	unlink(tcfgnm);
	return NULL;
    }
    close(tfilefd);
    write_XF86Config(state, tcfgnm);
    if(state->xf4driver != NULL){
        xf4file=malloc(strlen(tcfgnm)+10);
        strcpy(xf4file,tcfgnm);
        strcat(xf4file,"-4");
    }
    sync();
    if(state->xf4driver==NULL){
        snprintf(buf,sizeof(buf),"/etc/X11/X :9 -probeonly -verbose -pn -config %s "
                 " -bpp %d 2>> %s", tcfgnm, bpp, tfilenm1);
    }
    else{
        snprintf(buf,sizeof(buf),"/usr/X11R6/bin/Xorg :9 -probeonly -verbose -pn -config %s "
                 " -depth %d 2>> %s", xf4file, bpp, tfilenm1);
    }
    if (system(buf)) {
	system("/sbin/setsysfont 2> /dev/null");
	if (!kickstart)
	    fprintf(stderr, _("Unable to probe card - error running X\n"));
	unlink(tcfgnm);
	unlink(tfilenm1);
        if(xf4file)
            unlink(xf4file);
	return NULL;
    }
    system("/sbin/setsysfont 2> /dev/null");

    /* this is old way, doesnt work on Mach64, among others */
    /* new way, should be more robust */
    if (MaxValidMode(tfilenm1, &maxx, &maxy)) {
	unlink(tcfgnm);
	unlink(tfilenm1);
        if(xf4file)
            unlink(xf4file);
	return NULL;
    } else {
	mode = malloc(25);
	snprintf(mode, 25, "\"%dx%d\"", maxx, maxy);
    }

    /* all done with autoprobe information */
    unlink(tcfgnm);
    unlink(tfilenm1);
    if(xf4file)
        unlink(xf4file);

    if (!mode) {
	if (!kickstart && 0)
	    fprintf(stderr, _("Couldn't grep mode lines from probe output.\n"));
	return NULL;
    } else {
	char *s, *t;

	for (s = mode; *s && *s != '"'; s++);
	if (*s)
	    for (t = s+1; *t && *t != '"'; t++);
	else {
	    if (!kickstart && 0)
		fprintf(stderr, _("Couldn't grep mode lines from "
			"probe output.\n"));
	    free(mode);
	    return NULL;
	}
	if (*t)
	    *(t+1) = 0;
	else {
	    if (!kickstart && 0)
		fprintf(stderr, _("Couldn't grep mode lines from "
			"probe output.\n"));
	    free(mode);
	    return NULL;
	}
	rc=strdup(s);
	free(mode);
    }
    return rc;
}

/* 0 means everything went ok                                       */
/* < 0 means we blew it, bail configurations completely             */
/* > 0 means probing failed, user should configure manually         */
int kickstart_screen_configuration(struct XconfigState *state) {
    FILE *f;
    char *buf;
    int x8, y8, x16, y16, x32, y32;
    int disp8, disp16, disp32;
    int use8, use16, use32;
    int listed_doprobe, unlisted_doprobe;
    int rc;
    char tfilenm1[]="/tmp/Xprobe1-XXXXXX";
    char tfilenm2[]="/tmp/Xprobe2-XXXXXX";
    char tfilenm3[]="/tmp/Xprobe3-XXXXXX";
    char tcfgnm[]="/tmp/xf86config-XXXXXX";
    char *xf4file=NULL;
    char cmd[200];
    int tfilefd;
    char vmode[120], *s;
    int bestbpp, bestx, besty;

    /* run the selected kickstart card in probeonly mode */
    /* we'll try to pick out the ram size                */

    /* can also get here if user is ambitious and wants */
    /* us to autodetect modes for them                  */

    /*
     * Handle the Ramdac/Clockchip setting.
     * ******* For kickstart we just skip this *************
     */

    if (card_selected != UNLISTED_CARD) {
	if (card[card_selected].ramdac != NULL)
	    state->ramdac = card[card_selected].ramdac;
    } else {
	state->ramdac = NULL;
    }

    if (card_selected != UNLISTED_CARD) {
	state->clockchip = card[card_selected].clockchip;
    } else {
	state->clockchip = NULL;
    }

    /* throw screwy line in for modelines so X -probeonly will run ok */
    state->modesline8bpp = "\"640x480\"";
    state->modesline16bpp = state->modesline32bpp = NULL;

    /* just throw in some names */
    if (pciprobedcard)
	state->deviceidentifier = pciprobedcard;
    else
	state->deviceidentifier = "My Video Card";

    state->devicevendorname = "Unknown";
    state->deviceboardname = "Unknown";

    /* create all the temp files we'll need */
    if (!(tfilefd=mkstemp(tcfgnm))) {
	fprintf(stderr, _("Failed to create tempfile in kickstart mode\n"));
	exit(2);
    }
    close(tfilefd);
    if (!(tfilefd=mkstemp(tfilenm1))) {
	fprintf(stderr, _("Failed to create tempfile in kickstart mode\n"));
	unlink(tcfgnm);
	exit(2);
    }
    close(tfilefd);
    if (!(tfilefd=mkstemp(tfilenm2))) {
	fprintf(stderr, _("Failed to create tempfile in kickstart mode\n"));
	unlink(tcfgnm);
	unlink(tfilenm1);
	exit(2);
    }
    close(tfilefd);
    if (!(tfilefd=mkstemp(tfilenm3))) {
	fprintf(stderr, _("Failed to create tempfile in kickstart mode\n"));
	unlink(tcfgnm);
	unlink(tfilenm1);
	unlink(tfilenm2);
	exit(2);
    }
    close(tfilefd);

    /*
     * Run X -probeonly to figure out stuff about your card
     *
     * Write temporary XF86Config and run X -probeonly.
     */
    write_XF86Config(state, tcfgnm);
    sync();


    /* warn user something strange is going to happen */
    if (!kickstart)
    newtWinMessage(_("Probing to begin"), _("Ok"),
		   _("XorgTest will now run the X server you selected "
		     "to probe various information about your video card. "
		     "It is normal for the screen to blink several times."));
    if(state->xf4driver == NULL){
        snprintf(cmd, sizeof(cmd),
                 "/etc/X11/X :9 -probeonly -verbose -pn -config %s 2>> %s",
                 tcfgnm, tfilenm1);
    }
    else{
        snprintf(cmd, sizeof(cmd),
                 "/usr/X11R6/bin/Xorg :9 -probeonly -verbose -pn -config %s-4 2>> %s",
                 tcfgnm, tfilenm1);
        xf4file=malloc(strlen(tcfgnm)+5);
        xf4file=strcpy(xf4file,tcfgnm);
        strcat(xf4file,"-4");
    }

    rc=system(cmd);
    if (rc) {
	system("/sbin/setsysfont 2> /dev/null");
	unlink(tcfgnm);
        if(xf4file)
            unlink(xf4file);
        unlink(tfilenm1);
	unlink(tfilenm2);
	unlink(tfilenm3);
	if (!kickstart) {
	    newtWinMessage(_("X Error"), _("Ok"),
			   _("There was an error executing "
			     "the X server in a probing mode. You should try "
			     "configuring the video card manually."));
	    return 1;
	} else {
	    exit(2); /* what can we do in kickstart? No 'ASK' option yet */
	}
    }
    system("/sbin/setsysfont 2> /dev/null");

    /* do we need to grab out the clocks lines? */
    buf = alloca(8192);
    state->numberofclockslines = 0;
    listed_doprobe = 0;
    unlisted_doprobe = 0;
    if (card_selected == UNLISTED_CARD) {
	unlisted_doprobe = strcmp(state->serverbuf,"I128") &&
	    strcmp(state->serverbuf,"S3") &&
	    strcmp(state->serverbuf,"S3V") &&
	    strcmp(state->serverbuf,"Mach64") &&
	    strcmp(state->deviceidentifier, "Intel Corporation|i740") &&
	    strncmp(state->deviceidentifier, "Intel Corporation|82810", 23);
    } else {
	if (!(card[card_selected].flags & NOCLOCKPROBE))
	    listed_doprobe = 1;
    }
    
    /* don't bother for XFree 4 */
    if(state->xf4driver){
        listed_doprobe = 0;
        unlisted_doprobe = 0;
    }
    
    
    if  (unlisted_doprobe || listed_doprobe) {
	/* Look for 'clocks:' (case sensitive). */
	snprintf(cmd, sizeof(cmd), "grep clocks\\: %s >> %s",
		 tfilenm1, tfilenm2);
	if (system(cmd)) {
	    unlink(tcfgnm);
	    unlink(tfilenm1);
	    unlink(tfilenm2);
	    unlink(tfilenm3);
	    if (!kickstart) {
		newtWinMessage(_("X Error"), _("Ok"),
			     _("There was an error probing "
			       "the clocks on your video card. You should try "
			       "configuring the video card manually."));
		return 1;
	    } else {
		exit(2);
	    }
	}

	f = fopen(tfilenm2, "r");
	/* Parse lines. */
	while (fgets(buf, 8192, f) != NULL) {
	    char *clks;
	    clks = strstr(buf, "clocks: ") + 8;
	    if (clks >= buf + 3 && strcmp(clks - 11, "num") == 0)
		/* Reject lines with 'numclocks:'. */
		continue;
	    if (clks >= buf + 8 && strcpy(clks - 14, "pixel ") == 0)
		/* Reject lines with 'pixel clocks:'. */
		continue;
	    clks[strlen(clks) - 1] = '\0';	/* Remove '\n'. */
	    state->clocksline[state->numberofclockslines] =
		malloc(strlen(clks) + 1);
	    strcpy(state->clocksline[state->numberofclockslines],
		   clks);
	    state->numberofclockslines++;
	}
	fclose(f);
	unlink(tfilenm2);
    }

    /* find RAM information */
    snprintf(cmd, sizeof(cmd), "grep -i \"video[ ]*ram:\" %s >> %s", tfilenm1, tfilenm3);
    if (system(cmd)) {
	snprintf(cmd, sizeof(cmd), "grep -i \"kB of S[DG]RAM\" %s >> %s", tfilenm1, tfilenm3);
	if (system(cmd)) {
	    snprintf(cmd, sizeof(cmd), "grep -i \"VESA VBE Total Mem:\" %s >> %s", tfilenm1, tfilenm3);
	  if (system(cmd)) {
	     snprintf(cmd, sizeof(cmd), "grep -i \"VideoRAM:\" %s >> %s",
		      tfilenm1, tfilenm3);
	    if (system(cmd)) {
		unlink(tcfgnm);
		unlink(tfilenm1);
		unlink(tfilenm2);
		unlink(tfilenm3);
		
		if (!kickstart) {
		    newtWinMessage(_("X Error"), _("Ok"),
				   _("There was an error detecting "
				     "the video ram on your card. You should try "
				     "configuring the video card manually."));
		    return 1;
		} else {
		    exit(2); /* what can we do in kickstart? No 'ASK' option yet */
		}
	     }
	  }
	}
    }

    
    f = fopen(tfilenm3, "r");
    if (fgets(buf, 8192, f)) {
	char *s, *t, *e;
	int vram;

	s = strrchr(buf, ':');
	for (s=s+1; *s && !isdigit(*s); s++);
	for (t=s; *t && isdigit(*t); t++);
	*t = '\0';
	vram = strtol(s, &e, 10);

	if (e != s && *e == 0 && vram > 0) {
	    /* fix up vram, sometimes VBE reports it funny */
	    if (vram == 8128)
		vram = 8192;
	    state->videomemory = vram;
	} else {
	    unlink(tcfgnm);
	    unlink(tfilenm1);
	    unlink(tfilenm2);
	    unlink(tfilenm3);
	    if (!kickstart) {
		newtWinMessage(_("X Error"), _("Ok"),
			     _("There was an error detecting "
			       "the video ram on your card. You should try "
			       "configuring the video card manually."));
		return 1;
	    } else {
		exit(2); /* what can we do in kickstart? No 'ASK' option yet */
	    }
	}
    } else {
	unlink(tcfgnm);
	unlink(tfilenm1);
	unlink(tfilenm2);
	unlink(tfilenm3);
	if (!kickstart) {
	    newtWinMessage(_("X Error"), _("Ok"),
			 _("There was an error detecting "
			   "the video ram on your card. You should try "
			   "configuring the video card manually."));
	    return 1;
	} else {
	    exit(2); /* what can we do in kickstart? No 'ASK' option yet */
	}
    }
    fclose(f);

    /* clean up our little mess */
    unlink(tcfgnm);
    unlink(tfilenm1);
    unlink(tfilenm2);
    unlink(tfilenm3);
    if(xf4file)
        unlink(xf4file);
    /* figure out good mode lines          */
    /* uses ton of stuff in global vars :( */
    if ((rc=find_all_legal_modes(state)))
	return rc;

    /*
     * For the mono and vga16 server, no further configuration is
     * required.
     */
    if (state->screentype == 1 || state->screentype == 2) {
       /* VGA16 doesn't do 16 or 32 bpp */
       if (state->screentype == 2) {
	  state->modesline16bpp = "";
	  state->modesline32bpp = "";
       }
       return 0;
    }

    /*
     * Configure the modes order.
     */

    /* if we can't find a valid mode at 8bpp we're screwed */
    x8 = x16 = x32 = y8 = y16 = y32 = 0;
    disp8 = disp16 = disp32 = 0;
    use8 = use16 = use32 = 0;
    if ((state->modesline8bpp = autoprobe_bpp(state, 8)))
	modeline_to_res(state->modesline8bpp, &x8, &y8);
    if ((state->modesline16bpp = autoprobe_bpp(state, 16)))
	modeline_to_res(state->modesline16bpp, &x16, &y16);
    if(!xf4file){
        if ((state->modesline32bpp = autoprobe_bpp(state, 32)))
            modeline_to_res(state->modesline32bpp, &x32, &y32);
    }
    else{ /* XFree 4 seems to call this 24 bpp*/
        if ((state->modesline32bpp = autoprobe_bpp(state, 24)))
            modeline_to_res(state->modesline32bpp, &x32, &y32);
    }
    /* figure out which bpp to use */
    /* rule out ones with not a single valid mode line */
    disp8  = use8  = state->modesline8bpp  != NULL;
    disp16 = use16 = state->modesline16bpp != NULL;
    disp32 = use32 = state->modesline32bpp != NULL;

    if (!disp8 && !disp16 && !disp32) {
	newtWinMessage(_("No Valid Modes"), _("Ok"),
		     _("No valid video modes were autodetected at 8, 16, or "
		       "24 bits per pixel. It is likely that the X server "
		       "for your video card did not return the information "
		       "in the way expected by this tool.  You can still "
		       "attempt to configure you video card manually."));
	return 1;
    }

    /* require 1024x768, no matter what bpp this requires,    */
    /* unless both 8 and 16 bpp have same max physical, use 16*/
    if (use32 && x32 >= 1024 && y32 >= 768)
	use8 = use16 = 0;
    else if (use16 && x16 >= 1024 && y16 >= 768)
	use8 = use32 = 0;
    else if (use8)
	use16 = use32 = 0;

    /* should really check if all failed, but if that happened */
    /* we shouldnt have got this far (i hope)                  */

    /* blow away double quotes */
    if (use8) {
	bestbpp = 8;
	bestx = x8;
	besty = y8;
	strcpy(vmode, state->modesline8bpp+1);
    } else if (use16) {
	bestbpp = 16;
	bestx = x16;
	besty = y16;
	strcpy(vmode, state->modesline16bpp+1);
    } else if (use32) {
	bestbpp = 32;
	bestx = x32;
	besty = y32;
	strcpy(vmode, state->modesline32bpp+1);
    }

    s = strrchr(vmode, '"');
    *s = '\0';
    /* if in kickstart we just set it up and go, otherwise ask if we
       should just take this */

    if (!kickstart)
       rc = newtWinTernary(_("Probing finished"), _("Use Default"),
		   _("Let Me Choose"), _("Back"),
		   _("XorgTest has successfully probed your video "
		   "card. The default video "
		   "mode will be:"
		   "\n\n   Color Depth: %s bits per pixel\n"
		   "   Resolution : %s\n\nDo you want to accept "
		   "this setting, or select for yourself?"),
		   ((use8) ? "8" : ((use16) ? "16" : "24/32")), vmode);

    if (rc == 3) /* back */
      return 1;

    if (kickstart || rc == 1) {
	if (use8) {
	    state->modesline16bpp = NULL;
	    state->modesline32bpp = NULL;
	} else if (use16) {
	    state->modesline8bpp = NULL;
	    state->modesline32bpp = NULL;
	} else if (use32) {
	    state->modesline8bpp = NULL;
	    state->modesline16bpp = NULL;
	}

	return 0;
    }


    /* ok they think they know best                     */
    /* gotta build nice table of modes and color depths */
    /* let the user select the one they want, but we    */
    /* highlight the one we thought was best            */

    /*
     * Configure the modes order.
     */

    /* get list of possible modes again */
    /* populates state->modesline[8,16,32]bpp with modes that */
    /* should work in detected amount of ram                  */
    find_legal_modes(state);

    if (chooseXmodes(state, &state->modesline8bpp, &state->modesline16bpp,
		     &state->modesline32bpp) == 1)
     return 1;


    return 0;
}

void cpfailed(int foo) {
    char buf[2000];

    strcpy(buf, _("An error occured during probing;  please start over.\n\n"));

    if (foo == 1)
	strcat(buf, _("The error occured trying to exec X."));
    if (foo == 2)
	strcat(buf, _("The error occured trying to exec grep."));
    newtWinMessage(_("Error"), _("Ok"), buf);
    newtFinished();
    exit(2);
}

/*
 * Create the XF86Config file.
 */

static char *XF86firstchunk_text =
"# File generated by XConfigurator.\n"
"# **********************************************************************\n"
"\n"
"# **********************************************************************\n"
"# Module section -- this  section  is used to specify\n"
"# which dynamically loadable modules to load.\n"
"# **********************************************************************\n"
"\n"
"Section \"Module\"\n"
"\n"
"# This loads the DBE extension module.\n"
"    Load        \"dbe\"  	# Double buffer extension\n"
"# This loads the miscellaneous extensions module, and disables\n"
"# initialisation of the XFree86-DGA extension within that module.\n"
"    SubSection  \"extmod\"\n"
"      Option    \"omit xfree86-dga\"   # don't initialise the DGA extension\n"
"    EndSubSection\n"
"# This loads the font modules\n"
"    Load        \"type1\"\n"
"    #Load        \"speedo\"\n"
"    Load        \"freetype\"\n"
"    #Load        \"xtt\"\n"
"# This loads the GLX module\n"
"    Load        \"glx\"\n"
"# This loads the DRI module\n"
"    #Load        \"dri\"\n\n"
"EndSection\n"
"\n";
static char *XF86Config_firstchunk_text_sun =
"# **********************************************************************\n"
"# For xsun X servers, XF86Config can only contain a Files and Keyboard  \n"
"# section.\n"
"# **********************************************************************\n"
"\n";
static char *XF86Config_secondchunk_text =
"# **********************************************************************\n"
"# Files section.  This allows default font and rgb paths to be set\n"
"# **********************************************************************\n"
"\n"
"Section \"Files\"\n\n"
"    RgbPath	\"/usr/X11R6/lib/X11/rgb\"\n\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/misc/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/TTF/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/Type1/\"\n"
"    #FontPath   \"/usr/X11R6/lib/X11/fonts/CID/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/75dpi/\"\n"
"    FontPath  \"/usr/X11R6/lib/X11/fonts/100dpi/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/local/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/Speedo/\"\n"
"    FontPath   \"/usr/X11R6/lib/X11/fonts/cyrillic/\"\n\n"
"EndSection\n"
"\n";
static char *XF86Config_thirdchunk_text =
"# **********************************************************************\n"
"# Server flags section\n"
"# **********************************************************************\n"
"\n"
"Section \"ServerFlags\"\n"
"#    Option	\"NoTrapSignals\"\n"
"#    Option	\"DontZap\"\n"
"#    Option	\"DontZoom\"\n"
"EndSection\n"
"\n";

static char *keyboardsection_start =
"# **********************************************************************\n"
"# Core keyboard's InputDevice section\n"
"# **********************************************************************\n\n"
"Section \"InputDevice\"\n"
"\n"
"    Identifier  \"Keyboard1\"\n"
"    Driver	\"kbd\"\n"
"\n"
"    # Option	\"Protocol\"    \"Auto\"\n"
"\n"
"    # when using XQUEUE, comment out the above line, and uncomment the\n"
"    # following line\n"
"    # Option	\"Protocol\"    \"Xqueue\"\n"
"\n"
"    Option	\"AutoRepeat\"  \"500 30\"\n"
"\n"
"    # Let the server do the NumLock processing.  This should only be \n"
"    # required when using pre-R6 clients\n"
"    #Option	\"ServerNumLock\"\n"
"\n"
"    # Specify which keyboard LEDs can be user-controlled (eg, with xset(1))\n"
"    #Option	\"Xleds\"      \"1 2 3\n"
"\n"
"    #To set the LeftAlt to Meta, RightAlt key to ModeShift, \n"
"    #RightCtl key to Compose, and ScrollLock key to ModeLock:\n"
"\n"
"#     Option    \"LeftAlt\"        \"Meta\"\n";

static char *keyboardsection_part2 =
"#     Option    \"ScrollLock\"      \"Compose\"\n"
"#     Option    \"RightCtl\"        \"Control\"\n"
"\n"
"# To disable the XKEYBOARD extension, uncomment XkbDisable.\n"
"\n"
"#    Option    \"XkbDisable\"\n"
"\n"
"# To customise the XKB settings to suit your keyboard, modify the\n"
"# lines below (which are the defaults).  For example, for a non-U.S.\n"
"# keyboard, you will probably want to use:\n"
"#    Option    \"XkbModel\"    \"pc102\"\n"
"# If you have a US Microsoft Natural keyboard, you can use:\n"
"#    Option    \"XkbModel\"    \"microsoft\"\n"
"#\n"
"# Then to change the language, change the Layout setting.\n"
"# For example, a german layout can be obtained with:\n"
"#   Option     \"XkbLayout\"   \"de\"\n"
"# or:\n"
"#    Option    \"XkbLayout\"   \"de\"\n"
"#    Option    \"XkbVariant\"  \"nodeadkeys\"\n"
"#\n"
"# If you'd like to switch the positions of your capslock and\n"
"# control keys, use:\n"
"#    Option    \"XkbOptions\"  \"ctrl:swapcaps\"\n"
"\n"
"# These are the default XKB settings for Xorg\n"
"#    Option    \"XkbRules\"    \"xorg\"\n"
"#    Option    \"XkbModel\"    \"pc105\"\n"
"#    Option    \"XkbLayout\"   \"us\"\n"
"#    Option    \"XkbVariant\"  \"\"\n"
"#    Option   \" XkbOptions\"  \"\"\n"
"\n"
"    Option    \"XkbRules\"        \"xorg\"\n";

static char *keyboardsection_end =
"\n"
"EndSection\n\n";

#ifdef __sparc__
static char *keyboardsection_sun_start =
"# **********************************************************************\n"
"# Keyboard section\n"
"# **********************************************************************\n"
"\n"
"Section \"Keyboard\"\n"
"\n"
"    Protocol    \"Standard\"\n"
"\n"
"    # when using XQUEUE, comment out the above line, and uncomment the\n"
"    # following line (currently unsupported by xsun* servers)\n"
"    #Protocol   \"Xqueue\"\n"
"\n"
"    # AutoRepeat  500 5\n"
"\n"
"    # Let the server do the NumLock processing.  This should only be \n"
"    # required when using pre-R6 clients\n"
"    #ServerNumLock\n"
"\n"
"    # Specify which keyboard LEDs can be user-controlled (eg, with xset(1))\n"
"    #Xleds      1 2 3\n"
"\n"
"    #To set the LeftAlt to Meta, RightAlt key to ModeShift, \n"
"    #RightCtl key to Compose, and ScrollLock key to ModeLock:\n"
"\n";
static char *keyboardsection_sun_part2 =
"    LeftAlt        Meta\n"
"    RightAlt	    Meta\n"
"    ScrollLock     Compose\n"
"    RightCtl       Control\n"
"\n";
static char *keyboardsection_sun_part3 =
"# To disable the XKEYBOARD extension, uncomment XkbDisable.\n"
"\n"
"#    XkbDisable\n"
"\n"
"# These are the default XKB settings for Xorg\n"
"#    XkbRules    \"sun\"\n"
"#    XkbModel    \"type5\"\n"
"#    XkbLayout   \"us\"\n"
"#    XkbCompat   \"compat/complete\"\n"
"#    XkbTypes    \"types/complete\"\n"
"#    XkbKeycodes \"sun(type5)\"\n"
"#    XkbGeometry \"sun(type5)\"\n"
"#    XkbSymbols  \"sun/us(sun5)\"\n"
"\n"
"# These are the XKB settings determined by XorgTest\n"
"# and/or changed by the administrator\n";
static char *keyboardsection_sun_part4 =
"    XkbRules    \"sun\"\n"
"    XkbModel    \"%s\"\n"
"    XkbLayout   \"%s\"\n"
"    XkbCompat   \"compat/complete\"\n"
"    XkbTypes    \"types/complete\"\n"
"    XkbKeycodes \"%s\"\n"
"    XkbGeometry \"%s\"\n"
"    XkbSymbols  \"%s\"\n"
"\n"
"EndSection\n\n";
static char *keyboardsection_sun_part5 =
"# For xsun* servers, which does autodetection itself, it is safe\n"
"# to leave these commented out.\n"
"#   XkbRules    \"sun\"\n"
"#   XkbModel    \"%s\"\n"
"#   XkbLayout   \"%s\"\n"
"#   XkbCompat   \"compat/complete\"\n"
"#   XkbTypes    \"types/complete\"\n"
"#   XkbKeycodes \"%s\"\n"
"#   XkbGeometry \"%s\"\n"
"    XkbSymbols  \"%s\"\n"
"\n"
"EndSection\n\n";
#endif

static char *pointersection_text1 =
"# **********************************************************************\n"
"# Pointer section\n"
"# **********************************************************************\n"
"\n"
"Section \"InputDevice\"\n\n"
"    Identifier   \"Mouse1\"\n"
"    Driver       \"mouse\"\n";

static char *pointersection_text2 =
"\n"
"# When using XQUEUE, comment out the above two lines, and uncomment\n"
"# the following line.\n"
"\n"
"     #Option     \"Protocol\"	\"Xqueue\"\n"
"\n"
"# Mouse-speed setting for PS/2 mouse.\n"
"\n"
"     #Option     \"Resolution\"	\"256\"\n"
"\n"
"# Baudrate and SampleRate are only for some Logitech mice\n"
"\n"
"     #Option     \"BaudRate\"	9600\n"
"     #Option     \"SampleRate\"	150\n"
"\n"
"# Emulate3Buttons is an option for 2-button Microsoft mice\n"
"# Emulate3Timeout is the timeout in milliseconds (default is 50ms)\n"
"\n";

static char *monitorsection_text1 =
"# **********************************************************************\n"
"# Monitor section\n"
"# **********************************************************************\n"
"\n"
"# Any number of monitor sections may be present\n"
"\n"
"Section \"Monitor\"\n"
"\n";

static char *monitorsection_text2 =
"# HorizSync is in kHz unless units are specified.\n"
"# HorizSync may be a comma separated list of discrete values, or a\n"
"# comma separated list of ranges of values.\n"
"# NOTE: THE VALUES HERE ARE EXAMPLES ONLY.  REFER TO YOUR MONITOR\'S\n"
"# USER MANUAL FOR THE CORRECT NUMBERS.\n"
"\n";

static char *monitorsection_text3 =
"# VertRefresh is in Hz unless units are specified.\n"
"# VertRefresh may be a comma separated list of discrete values, or a\n"
"# comma separated list of ranges of values.\n"
"# NOTE: THE VALUES HERE ARE EXAMPLES ONLY.  REFER TO YOUR MONITOR\'S\n"
"# USER MANUAL FOR THE CORRECT NUMBERS.\n"
"\n";

static char *monitorsection_text4 =
"# Modes can be specified using the format below, if needed.\n"
"\n"
"#    ModeLine \"1024x768i\" 45 1024 1048 1208 1264 768 776 784 817 Interlace\n"
"\n";

static char *modelines_text_Trident_TG_96xx =
"# This is a set of standard mode timings. Modes that are out of monitor spec\n"
"# are automatically deleted by the server (provided the HorizSync and\n"
"# VertRefresh lines are correct), so there's no immediate need to\n"
"# delete mode timings (unless particular mode timings don't work on your\n"
"# monitor). With these modes, the best standard mode that your monitor\n"
"# and video card can support for a given resolution is automatically\n"
"# used.\n"
"\n"
"# These are special modelines for Trident Providia 9685. It is for VA Linux\n"
"# systems only.\n"
"# 640x480 @ 72 Hz, 36.5 kHz hsync\n"
"Modeline \"640x480\"     31.5   640  680  720  864   480  488  491  521\n"
"# 800x600 @ 72 Hz, 48.0 kHz hsync\n"
"Modeline \"800x600\"     50     800  856  976 1040   600  637  643  666 +hsync +vsync\n"
"# 1024x768 @ 60 Hz, 48.4 kHz hsync\n"
"#Modeline \"1024x768\"    65    1024 1032 1176 1344   768  771  777  806 -hsync -vsync\n"
"# 1024x768 @ 70 Hz, 56.5 kHz hsync\n"
"Modeline \"1024x768\"    75    1024 1048 1184 1328   768  771  777  806 -hsync -vsync\n"
"\n";

static char *modelines_text_MediaGX =
"# This is a set of standard mode timings. Modes that are out of monitor spec\n"
"# are automatically deleted by the server (provided the HorizSync and\n"
"# VertRefresh lines are correct), so there's no immediate need to\n"
"# delete mode timings (unless particular mode timings don't work on your\n"
"# monitor). With these modes, the best standard mode that your monitor\n"
"# and video card can support for a given resolution is automatically\n"
"# used.\n"
"\n"
"# These are special modelines for the Cyrix MediaGX\n"
"\n"
"Modeline \"640x400\"     25.175 640  656  760  800   400  409  411  450\n"
"# 640x480 @ 60 Hz, 31.5 kHz hsync\n"
"Modeline \"640x480\"     25.175 640  656  760  800   480  491  493  525\n"
"# 800x600 @ 56 Hz, 35.15 kHz hsync\n"
"ModeLine \"800x600\"     36     800  816  896 1024   600  601  603  625\n"
"# 640x400 @ 85 Hz, 37.86 kHz hsync\n"
"Modeline \"640x400\"     31.5   640  656 736   832   400  401  404  445 -HSync +VSync\n"
"# 640x480 @ 72 Hz, 36.5 kHz hsync\n"
"Modeline \"640x480\"     31.5   640  656  720  864   480  488  491  521\n"
"# 640x480 @ 75 Hz, 37.50 kHz hsync\n"
"ModeLine  \"640x480\"    31.5   640  656  720  840   480  481  484  500 -HSync -VSync\n"
"# 800x600 @ 60 Hz, 37.8 kHz hsync\n"
"Modeline \"800x600\"     40     800  816  968 1056   600  601  605  628 +hsync +vsync\n"
"\n"
"# 640x480 @ 85 Hz, 43.27 kHz hsync\n"
"Modeline \"640x480\"     36     640  656  752  832   480  481  484  509 -HSync -VSync\n"
"# 800x600 @ 72 Hz, 48.0 kHz hsync\n"
"Modeline \"800x600\"     50     800  816  976 1040   600  637  643  666 +hsync +vsync\n"
"# 1024x768 @ 60 Hz, 48.4 kHz hsync\n"
"Modeline \"1024x768\"    65    1024 1040 1176 1344   768  771  777  806 -hsync -vsync\n"
"\n"
"# 640x480 @ 100 Hz, 53.01 kHz hsync\n"
"Modeline \"640x480\"     45.8   640  656  768  864   480  488  494  530 -HSync -VSync\n"
"# 800x600 @ 85 Hz, 55.84 kHz hsync\n"
"Modeline  \"800x600\"    60.75  800  816  928 1088   600  616  621  657 -HSync -VSync\n"
"\n"
"# 1024x768 @ 70 Hz, 56.5 kHz hsync\n"
"Modeline \"1024x768\"    75    1024 1040 1184 1328   768  771  777  806 -hsync -vsync\n"
"# 800x600 @ 100 Hz, 64.02 kHz hsync\n"
"Modeline  \"800x600\"    69.65  800  816  928 1088   600  604  610  640 -HSync -VSync\n"
"# 1024x768 @ 76 Hz, 62.5 kHz hsync\n"
"Modeline \"1024x768\"    85    1024 1040 1152 1360   768  784  787  823\n"
"\n"
"# 1024x768 @ 85 Hz, 70.24 kHz hsync\n"
"Modeline \"1024x768\"   98.9  1024 1040 1216 1408   768 782 788 822 -HSync -VSync\n"
"\n"
"# 1024x768 @ 100Hz, 80.21 kHz hsync\n"
"Modeline \"1024x768\"   115.5  1024 1040 1248 1440  768  771  781  802 -HSync -VSync\n"
"\n";

static char *devicesection_text =
"# **********************************************************************\n"
"# Graphics device section\n"
"# **********************************************************************\n"
"\n"
"# Any number of graphics device sections may be present\n"
"\n"
"Section \"Device\"\n"
"    Identifier       \"Standard VGA\"\n"
"    VendorName       \"Unknown\"\n"
"    BoardName        \"Unknown\"\n"
"    #Chipset         \"generic\"\n"
"\n"
"    Driver           \"vga\"\n"
"\n"
"    #VideoRam        \"256\"\n"
"\n"
"    #Clocks          \"25.2 28.3\"\n"
"\n"
"EndSection\n"
"\n"
"# Device configured by XorgTest:\n"
"\n";

static char *screensection_text1 =
"# **********************************************************************\n"
"# Screen sections\n"
"# **********************************************************************\n"
"\n";

static char *serversection_text =
"# **********************************************************************\n"
"# ServerLayout sections.\n"
"# **********************************************************************\n"
"\n"
"# Any number of ServerLayout sections may be present.\n"
"\n"
"Section \"ServerLayout\"\n"
"    Identifier   \"Simple Layout\"\n"
"    Screen       \"Screen 1\"\n"
"    InputDevice  \"Mouse1\" \"CorePointer\"\n"
"    InputDevice  \"Keyboard1\" \"CoreKeyboard\"\n"
"\n"
"EndSection\n"
"\n";

static int is_Riva128(const char *card)
{
    return !strncmp(card, "Riva 128", 8)
	   || !strncmp(card, "Nvidia Corporation|Riva TNT 128", 31)
	   || !strncmp(card, "NVidia / SGS Thomson|Riva128", 28);
}

void write_XF86Config(struct XconfigState *state, char *filename) {

    FILE *f;
    char *xf4filename;
    GSList *temp;
    int defdepth=0;
    
    
    newtComponent form, ok;

    /* Write the configuration file for XFree4 first if we have a driver */

    if(state->xf4driver!=NULL){
        xf4filename=(char*)malloc(strlen(filename)+10);
        strcpy(xf4filename,filename);
        strcat(xf4filename,"-4");
        write_XF86Config4(state,xf4filename);
        free(xf4filename);
    }
    
    /* Write the file. */
                         
    f = fopen(filename, "w");
    if (f == NULL) {
	if (getuid() != 0)

	    newtOpenWindow(10, 2, 60, 20, "Error!");
	form = newtForm(NULL, NULL, 0);
	newtFormAddComponent(form, newtLabel(3, 1,
					   _("Can't open file for writing.  "
					     "Perhaps you aren't root?")));
	ok = newtButton(3, 25, _("Ok"));
	newtFormAddComponents(form, ok, NULL);
	newtRunForm(form);
	newtFormDestroy(form);
	newtPopWindow();
	newtFinished();

	exit(-1);
    }

    fprintf(f, "%s", XF86firstchunk_text);
    if (state->screentype == 5)
	fprintf(f, "%s", XF86Config_firstchunk_text_sun);
    fprintf(f, "%s", XF86Config_secondchunk_text);
    if (state->screentype <= 4)
	fprintf(f, "%s", XF86Config_thirdchunk_text);

    /*
     * Write keyboard section.
     */
#ifdef __sparc__
    if (state->xkb_rules && !strcmp (state->xkb_rules, "sun")) {
	fprintf(f, "%s", keyboardsection_sun_start);
	if (state->screentype != 5)
	    fprintf(f, "%s", keyboardsection_sun_part2);
	fprintf(f, keyboardsection_sun_part3);
	fprintf(f, state->screentype == 5 ? keyboardsection_sun_part5 : keyboardsection_sun_part4,
		state->xkb_model, state->xkb_layout,
		state->xkb_keycodes, state->xkb_geometry, state->xkb_symbols);
     } else {
	 strcpy(state->mousetype, "PS/2");
#endif
    fprintf(f, "%s", keyboardsection_start);

    if (state->altmeta)
	fprintf(f, "    #Option    \"RightAlt\"       \"ModeShift\"\n");
    else
	fprintf(f, "    #Option    \"RightAlt\"        \"Meta\"\n");

    fprintf(f, "%s", keyboardsection_part2);
    fprintf(f, "    Option    \"XkbModel\"        \"%s\"\n", strcmp(state->xkb_keymap, "jp") ? "pc105" : "jp106");
    fprintf(f, "    Option    \"XkbLayout\"       \"%s\"\n", state->xkb_keymap);
    fprintf(f, "%s", keyboardsection_end);
#ifdef __sparc__
    }
#endif

    if (state->screentype == 5) {
      fclose(f);
      return;
    }

    /*
     * Write pointer section.
     */
    fprintf(f, "%s", pointersection_text1);
    fprintf(f, "    Option    \"Protocol\"    \"%s\"\n", state->mousetype);
    fprintf(f, "    Option    \"Device\"      \"%s\"\n", state->pointerdevice);

    /* ZAxisMapping and Emulate3Buttons clash on XFree 3.3 */
    
    if(!state->emulate3buttons)
        fprintf(f, "    Option    \"ZAxisMapping 4 5 6 7\"\n");
    fprintf(f, "%s", pointersection_text2);
    if (!state->emulate3buttons)
	fprintf(f, "#");
    fprintf(f, "    Option    \"Emulate3Buttons\"\n");
    if (!state->emulate3buttons)
	fprintf(f, "#");
    fprintf(f, "    Option    \"Emulate3Timeout\"    50\"\n\n");
    fprintf(f, "# ChordMiddle is an option for some 3-button "
	    "Logitech mice\n\n");
    if (!state->chordmiddle)
	fprintf(f, "#");
    fprintf(f, "    Option    \"ChordMiddle\"\n\n");
    if (state->cleardtrrts) {
	fprintf(f, "    Option    \"ClearDTR\"\n");
	fprintf(f, "   Option    \"ClearRTS\"\n\n");
    }
    fprintf(f, "EndSection\n\n");

    /*
     * Write monitor section.
     */
    fprintf(f, "%s", monitorsection_text1);
    fprintf(f, "    Identifier  \"%s\"\n", state->monitoridentifier);
    fprintf(f, "    VendorName  \"%s\"\n", state->monitorvendorname);
    fprintf(f, "    ModelName   \"%s\"\n", state->monitormodelname);
    fprintf(f, "\n");
    fprintf(f, "%s", monitorsection_text2);
    fprintf(f, "    HorizSync   %s\n", state->hsyncrange);
    fprintf(f, "\n");
    fprintf(f, "%s", monitorsection_text3);
    fprintf(f, "    VertRefresh %s\n", state->vsyncrange);
    fprintf(f, "\n");
    fprintf(f, "%s", monitorsection_text4);
    if (pciprobedcard
        && combo){
        fprintf(f,"#Specific modelines for this model\n");
        temp=combo->modelines;
        while(temp!=NULL){
            fprintf(f,"%s\n",(char*)(temp->data));
            temp=g_slist_next(temp);
                
        }
    }
    else if (pciprobedcard
        && (strncasecmp(pciprobedcard, "TG 96", 5) == 0
            || strncasecmp(pciprobedcard, "Trident Microsystems|TGUI 96", 28) == 0)){
	fprintf(f, "%s", modelines_text_Trident_TG_96xx);
    }
    else{
        if(card_selected != UNLISTED_CARD && strstr(card[card_selected].chipset, "MediaGX")){
            fprintf(f, "%s", modelines_text_MediaGX);
        }
        /* else
            fprintf(f, "%s", modelines_text); */
    }
    fprintf(f, "EndSection\n\n\n");

    /*
     * Write Device section.
     */

    fprintf(f, "%s", devicesection_text);
    fprintf(f, "Section \"Device\"\n");
    fprintf(f, "    Identifier  \"%s\"\n", state->deviceidentifier);
    fprintf(f, "    Driver      \"%s\"\n", state->xf4driver);
    fprintf(f, "    VendorName  \"%s\"\n", state->devicevendorname);
    fprintf(f, "    BoardName   \"%s\"\n", state->deviceboardname);

    /* Riva 128 can't probe video ram */
    if (pciprobedcard && (is_Riva128(pciprobedcard)))
        fprintf(f, "    VideoRam    %d\n", state->videomemory);
    else
        /* Rely on server to detect video memory. */
        fprintf(f, "    #VideoRam    %d\n", state->videomemory);

    if (card_selected != UNLISTED_CARD &&
	strcmp(state->deviceidentifier,"NeoMagic 128XD (laptop/notebook)"))
	/* Add comment lines from card definition. */
	fprintf(f, card[card_selected].lines);
    if (state->ramdac != NULL && *state->ramdac != '\0')
	fprintf(f, "    Ramdac      \"%s\"\n", state->ramdac);
    if (card_selected != UNLISTED_CARD)
	if (card[card_selected].dacspeed != NULL)
	    fprintf(f, "    Dacspeed    %s\n",
		    card[card_selected].dacspeed);
    if (state->clockchip != NULL)
	fprintf(f, "    Clockchip   \"%s\"\n", state->clockchip);
    else
	if (state->numberofclockslines == 0)
	    fprintf(f, "    # Insert Clocks lines here if appropriate\n");
	else {
	    int i;
	    for (i = 0; i < state->numberofclockslines; i++)
		fprintf(f, "    Clocks %s\n", state->clocksline[i]);
	}
    fprintf(f, "EndSection\n\n");


    /*
     * Find the default colour depth.  Default to 16, then 24/32, finally 8.
     */

    if(state->modesline8bpp && strcmp(state->modesline8bpp, ""))
        defdepth=8;
    if(defdepth==0 && state->modesline32bpp && strcmp(state->modesline32bpp, ""))
        defdepth=32;
    if(state->modesline16bpp && strcmp(state->modesline16bpp, ""))
        defdepth=16;
    
    if (!strncmp(state->deviceidentifier,"Silicon Motion",14)  && defdepth == 32) {	
	    defdepth = 24;
    }
    
    /*
     * Write Screen sections.
     */

    fprintf(f, "%s", screensection_text1);

    /*
     * The Accel section.
     */
    fprintf(f,

	    "# The accelerated server\n"
	    "Section \"Screen\"\n"
	    "    Identifier  \"Screen1\"\n"
	    "    Device      \"%s\"\n"
	    "    Monitor     \"%s\"\n", state->deviceidentifier,
	    state->monitoridentifier);
    if(defdepth!=0){
        fprintf(f,
                "    DefaultDepth   %i\n\n",defdepth);
    }

    if (state->modesline8bpp && strcmp(state->modesline8bpp, ""))
	fprintf(f,
		"    Subsection \"Display\"\n"
		"        Depth       8\n"
		"        Modes       %s\n"
		"        ViewPort    0 0\n"
		"    EndSubsection\n", state->modesline8bpp);
    if (state->modesline16bpp && strcmp(state->modesline16bpp, ""))
	fprintf(f,
		"    Subsection \"Display\"\n"
		"        Depth       16\n"
		"        Modes       %s\n"
		"        ViewPort    0 0\n"
		"    EndSubsection\n", state->modesline16bpp);
    if (state->modesline32bpp && strcmp(state->modesline32bpp, ""))
	fprintf(f,
		"    Subsection \"Display\"\n"
		"        Depth       24\n"
		"        Modes       %s\n"
		"        ViewPort    0 0\n"
		"    EndSubsection\n",
		state->modesline32bpp);
    if (state->modesline32bpp && strcmp(state->modesline32bpp, ""))
	fprintf(f,
		"    Subsection \"Display\"\n"
		"        Depth       32\n"
		"        Modes       %s\n"
		"        ViewPort    0 0\n"
		"    EndSubsection\n",
		state->modesline32bpp);
    fprintf(f, "EndSection\n\n");

    fprintf(f, "%s", serversection_text);

    fclose(f);
}

/*
  Writes an XFree86 config file for XFree 4
*/

void write_XF86Config4(struct XconfigState *state, char *filename) {
  FILE *f;
  newtComponent form, ok;
  int preferredColorDepth=8;
  GSList *temp;
  
  
  /*
   * Write the file.
   */
 
  f = fopen(filename, "w");
  if (f == NULL) {
    if (getuid() != 0)
      newtOpenWindow(10, 2, 60, 20, "Error!");
    form = newtForm(NULL, NULL, 0);
    newtFormAddComponent(form, newtLabel(3, 1,
					 _("Can't open file for writing.  "
					   "Perhaps you aren't root?")));
    ok = newtButton(3, 25, _("Ok"));
    newtFormAddComponents(form, ok, NULL);
    newtRunForm(form);
    newtFormDestroy(form);
    newtPopWindow();
    newtFinished();
    
    exit(-1);
  }
  /*      Header... */

  fprintf(f,"# Xorg configuration generated by XorgTest %s\n\n", VERSION);

  /* A generic server layout header */

  fprintf(f,"Section \"ServerLayout\"\n");
  fprintf(f,"\tIdentifier \"Xorg Configured\"\n");
  fprintf(f,"\tScreen      0  \"Screen0\" 0 0\n");
  fprintf(f,"\tInputDevice    \"Mouse1\" \"CorePointer\"\n"); 
  fprintf(f,"\tInputDevice    \"Keyboard1\" \"CoreKeyboard\"\n");
  fprintf(f,"EndSection\n\n");

  /* Write the xfs path */
    
  fprintf(f,"# Red Hat Linux 6.0 and later use the xfs font server for improved\n");
  fprintf(f,"# interactive performance\n\n");
  fprintf(f,"Section \"Files\"\n");
  fprintf(f,"\tFontPath \"unix/:7100\"\n");
  fprintf(f,"EndSection\n\n");
  
  /* Load all modules */
  
  fprintf(f,"# Module loading section\n\n");
  fprintf(f,"Section \"Module\"\n");
  fprintf(f,"\tLoad  \"dbe\"\t\t# Double-buffering\n");
  fprintf(f,"\tLoad  \"GLcore\"\t\t# OpenGL support\n");
  fprintf(f,"%sLoad  \"dri\"\t\t# Direct rendering infrastructure\n",
          (nodri == 1) ? "\t# ":"\t" );	
  fprintf(f,"\tLoad  \"glx\"\t\t# OpenGL X protocol interface\n");
  fprintf(f,"\tLoad  \"extmod\"\t\t# Misc. required extensions\n");
  fprintf(f,"\tLoad  \"v4l\"\t\t# Video4Linux\n");
  fprintf(f,"\t# Load  \"record\"\t# X event recorder\n");
  fprintf(f,"\t# You only need the following two modules if you do not use xfs.\n");
  fprintf(f,"\t# Load  \"freetype\"\t # TrueType font handler\n");
  fprintf(f,"\t# Load  \"type1\"\t\t# Adobe Type 1 font handler\n");
  fprintf(f,"EndSection\n\n");

  /*
   * The keyboard section - assume all of them use the
   * keyboard driver
   */

  fprintf(f,"Section \"InputDevice\"\n");
  fprintf(f,"\tIdentifier  \"Keyboard1\"\n");
  fprintf(f,"\tDriver      \"keyboard\"\n");
  if(state->xkb_keymap)
      fprintf(f,"\tOption      \"XkbLayout\"\t\"%s\"\n",state->xkb_keymap);
  if(strcmp(state->xkb_keymap,"jp")==0)
      fprintf(f,"\tOption      \"XkbModel\"\t\"%s\"\n","jp106");
  fprintf(f,"EndSection\n\n");
    
  /*
   * Mouse support
   */

  fprintf(f,"Section \"InputDevice\"\n");
  fprintf(f,"\tIdentifier  \"Mouse1\"\n");
  fprintf(f,"\tDriver      \"mouse\"\n");
  fprintf(f,"\tOption      \"Device\" \"%s\"\n", state->pointerdevice);
  fprintf(f,"\tOption      \"Protocol\" \"%s\"\n",state->mousetype);
  fprintf(f,"\tOption      \"Emulate3Buttons\" \"%s\"\n",
            (state->emulate3buttons)?"on":"off");
  fprintf(f,"\tOption      \"ZAxisMapping\" \"4 5 6 7\"\n");
  fprintf(f,"EndSection\n\n");

  /*
   * Monitor information
   */

  fprintf(f,"Section \"Monitor\"\n");
  fprintf(f,"\tIdentifier \"%s\"\n",state->monitoridentifier); 
  fprintf(f,"\tVendorName \"%s\"\n",state->monitorvendorname);
  fprintf(f,"\tModelName  \"%s\"\n",state->monitormodelname);
  fprintf(f,"\tHorizSync %s\n",state->hsyncrange);
  fprintf(f,"\tVertRefresh %s\n",state->vsyncrange);
  if (state->dpms)  
      fprintf(f,"\tOption \"dpms\"\n");

/* If some modelines has been specified, write them */
  
  if (pciprobedcard && combo){
      fprintf(f,"# Specific modelines for this model\n");
      temp=combo->modelines;
      while(temp!=NULL){
          fprintf(f,"%s\n",(char*)(temp->data));
          temp=g_slist_next(temp);
          
      }
  }

  fprintf(f,"EndSection\n\n");


  /*
   * Driver information.
   */
     
  fprintf(f,"Section \"Device\"\n");
  fprintf(f,"\tIdentifier \"%s\"\n",state->deviceidentifier);
  fprintf(f,"\tDriver \"%s\"\n",state->xf4driver);
  fprintf(f,"\tBoardName \"%s\"\n",state->deviceboardname);
  if(usefbdev) {
      fprintf(f, "\tOption \"UseFBDev\"\n");
      fprintf(f, "\tOption \"NoHWCursor\"\n");
  }	  
  /*
   * If not set, i810 chokes
   */
  if(strcmp("i810",state->xf4driver)==0){
      fprintf(f,"\tVideoRam %i\n",state->videomemory);
  }
  if (card_selected != UNLISTED_CARD)
	/* Add comment lines from card definition. */
	fprintf(f, card[card_selected].lines);
  fprintf(f,"EndSection\n\n");
    

  /*
   * Frame buffer driver section.
   */
  
  fprintf(f,"Section \"Device\"\n");
  fprintf(f,"\tIdentifier \"Linux Frame Buffer\"\n");
  fprintf(f,"\tDriver \"fbdev\"\n");
  fprintf(f,"\tBoardName \"%s\"\n",state->deviceboardname);
  fprintf(f,"EndSection\n\n");
  
  /*
   * Screen section
   */

  fprintf(f,"Section \"Screen\"\n");
  fprintf(f,"\tIdentifier \"Screen0\"\n");
  fprintf(f,"\tDevice \"%s\"\n",state->deviceidentifier);
  fprintf(f,"\tMonitor \"%s\"\n",state->monitoridentifier);

  /* XFree 4 needs a DefaultDepth -
   * Prefer 16, then 24, finally 8
   */
  
  if (state->modesline8bpp && strcmp(state->modesline8bpp,""))
      preferredColorDepth = 8;
  if (state->modesline32bpp && strcmp(state->modesline32bpp,""))
      preferredColorDepth = 24;
  if (state->modesline16bpp && strcmp(state->modesline16bpp,""))
      preferredColorDepth = 16;
  fprintf(f,"\tDefaultDepth %i\n", preferredColorDepth);

/*
 * 24 bpp equals 32 bpp, I hope...
 */

  if (state->modesline32bpp && strcmp(state->modesline32bpp,"")){
      fprintf(f,"\tSubsection \"Display\"\n"
              "\t\tDepth 24\n"
              "\t\tModes %s\n"
              "\tEndSubSection\n",state->modesline32bpp);
  }

  if (state->modesline16bpp && strcmp(state->modesline16bpp,"")){
      fprintf(f,"\tSubsection \"Display\"\n"
              "\t\tDepth 16\n"
              "\t\tModes %s\n"
              "\tEndSubSection\n",state->modesline16bpp);
  }
      
  if (state->modesline8bpp && strcmp(state->modesline8bpp,"")){
      fprintf(f,"\tSubsection \"Display\"\n"
              "\t\tDepth 8\n"
              "\t\tModes %s\n"
              "\tEndSubSection\n",state->modesline8bpp);
  }
  
  fprintf(f,"EndSection\n\n");
    
  /* DRI section */
  
  fprintf(f,"Section \"DRI\"\n");
  fprintf(f,"\tMode 0666\n");
  fprintf(f,"EndSection\n"); 
 
  fclose(f);

}
    
    /* Ask where to write XF86Config to. Returns filename. */

char *ask_XF86location() {

    if (getuid() == 0) {
#ifdef PREFER_XORGIN_ETC
	return "/etc/X11/xorg.conf";
#endif
    }

    return "xorg.conf";
}


char ** splitString(char * str, int length, char sep) {
    char * s, * source, * dest;
    char ** list;
    int i;
    int fields;

    s = malloc(length + 1);

    fields = 1;
    for (source = str, dest = s, i = 0; i < length; i++, source++, dest++) {
        *dest = *source;
        if (*dest == sep) fields++;
    }

    *dest = '\0';

    list = malloc(sizeof(char *) * (fields + 1));

    dest = s;
    list[0] = dest;
    i = 1;
    while (i < fields) {
        if (*dest == sep) {
            list[i++] = dest + 1;
            *dest = 0;
        }
        dest++;
    }

    list[i] = NULL;

    return list;
}


void write_file(struct XconfigState *state)
{
    FILE *fd;

    fd = fopen("/var/run/SERVER", "w");
    
    fprintf(fd, "%s %d\n", (state->xf4driver == NULL) ? state->serverbuf
	    : "Xorg", card_selected);
    
    fclose(fd);
}

char *read_file()
{
    FILE *fd;
    char *buf;
    char cardbuf[10];
    int card;

    buf = malloc(sizeof(char) * 80);

    fd = fopen("/var/run/SERVER", "r");
    if (!fd) {
	fprintf(stderr, "You must run with --pick first\n");
	exit(1);
    }
    fscanf(fd, "%s %s\n", buf, cardbuf);
    fclose(fd);

    card = atoi(cardbuf);

    card_selected = card;
    unlink("/var/run/SERVER");
    return buf;

}


/* common problem - /usr/X11R6/lib/X11/xorg.conf is supposed */
/* to be a link to /etc/X11/xorg.conf. If you run a XFree86  */
/* configuration utility, this link can be destroyed.         */
/* Notify user if this is the case                            */
#define LINK_OK 255
#define LINK_FAIL 0
#define LINK_FATAL -1
#define LINK_FIXABLE 128

/* should we automatically probe cause we know the card can do it */
int DoAutoProbe(struct XconfigState *state) {
    /* ok so far I know these are ok:         */
    /*                                        */
    /* Servers:                               */
    /* --------                               */
    /*    S3                                  */
    /*   I128                                 */
    /*    S3V                                 */
    /*  Mach64                                */
    /*                                        */
    /* Cards:                                 */
    /* ------                                 */
    /*   Matrox Millennium (MGA)              */
    /*   Matrox Millennium II                 */
    /*   Matrox Mystique                      */

    if (card_selected != UNLISTED_CARD) {
	if (card[card_selected].server &&
	    (!strcmp(card[card_selected].server, "S3") ||
	     !strcmp(card[card_selected].server, "S3V") ||
	     strstr(card[card_selected].chipset, "S3 ViRGE") ||
	     !strcmp(card[card_selected].server, "I128") ||
	     !strcmp(card[card_selected].server, "Mach64")))
	    return 1;
	else if (!strcmp(card[card_selected].name, "Matrox Millennium (MGA)") ||
		 !strcmp(card[card_selected].name, "Matrox Millennium II") ||
		 !strcmp(card[card_selected].name, "Matrox Millennium II AGP") ||
		 !strcmp(card[card_selected].name, "Matrox Millennium G200") ||
		 !strcmp(card[card_selected].name, "Matrox Millennium G400") ||
		 !strcmp(card[card_selected].name, "Matrox Millennium G450") ||
		 !strcmp(card[card_selected].name, "Matrox Mystique"))
	    return 1;
    } else if (!strcmp(state->serverbuf, "XF86_S3") ||
	       !strcmp(state->serverbuf, "XF86_S3V") ||
	       !strcmp(state->serverbuf, "XF86_I128") ||
	       !strcmp(state->serverbuf, "XF86_Mach64")) {
	return 1;
    } else
	return 0;
    
    return 0; /* shouldnt get here! */
}

void usage(int rc) {
    fprintf(stderr,_("\nUsage: XorgTest [--kickstart] [--card <card>]\n"
            "       [--server <server>] [--driver <driver>]\n"
            "       [--monitor <monitor>] [--hsync <hsync>] [--vsync <vsync>]\n"
            "       [--mouse <mouse>] [--mousedevice <port>] [--emulthree]\n"
            "       [--startxonboot] [--preferxf3] [--preferxf4]\n"
            "       [--listservers] [--listmonitors] [--listcards]\n"
            "       [--usefbdev] [--noddcprobe]\n\n"));
    fprintf(stderr,_("       --kickstart - try to automatically probe card.\n\n"));
    fprintf(stderr,_("\n       Run 'mouseconfig --help' for info about "
            "--mouse and --mousedevice.\n"));
    fprintf(stderr,_("\n       Just one of --card and --server needs "
            "to be specified."
            " If neither given\n       the PCI bus will be scanned for "
            "a video card.\n\n"));
    fprintf(stderr,_("       Specify either --monitor, or --hsync and "
            "--vsync.\n"));
    fprintf(stderr,_("\n       Example: XorgTest --kickstart "
            "--monitor \"Cool Tube 1000\" \\\n                 "
            "--mouse ps/2 --mousedevice ttyS0\n"));
    fprintf(stderr,_("\n       If --kickstart is NOT specified, the user "
            "will be interactively queried\n"
            "       for card type and monitor type.\n\n"));
    exit(rc);
}

int screen_autoprobe(struct XconfigState *state)
{
  /* some of these guys hate to be poked               */
  /* if we dont know then its at the user's discretion */
  if (strncmp(state->serverbuf, "XF86_TGA", 3) &&
      strncmp(state->serverbuf, "XF86_Mach32", 6) &&
      (pciprobedcard ? !is_Riva128(pciprobedcard) : 1)) {
    if (DoAutoProbe(state) && !expert) {
	if (kickstart_screen_configuration(state) > 0) {
	    if (custom)
		return custommonitor_configuration(state, 1);
	    else
		return 1;
	}
    } else {
      if (newtWinChoice(_("Screen Configuration"),
			_("Don't Probe"), _("Probe"),
			_("XorgTest now needs to setup the default "
			  "resolution and color depth.  Most modern PCI "
			  "video "
			  "cards can be probed, and XorgTest will "
			  "automatically determine the best "
			  "video mode/color depth possible for your "
			  "system. There is a chance, however, this could "
			  "lock up your system.  If you would prefer "
			  "to give the required information instead of "
			  "having it probed, answer \"Don't Probe\" to "
			  "the following question.")) != 2) {
	return screen_configuration(state);
      } else {
	if (kickstart_screen_configuration(state) > 0) {
	    if (custom)
		return custommonitor_configuration(state, 1);
	    else
		return 1;
	}
      }
    }
  } else {
    return screen_configuration(state);
  }
  return 0;
}

/*
  Parses a syncline ("  abc.d - efgh.i ") into numbers
  Returns -1 and -1 on failure.
*/

void parseSyncLine(char *line,double *min, double *max)
{
  int len;
  char *loc;
  
  len=strlen(line);
  *min=strtod(line,&loc);
  do{
    loc++;
  }while(((loc[0] < '0') || (loc[0] > '9')) && (loc<(line+len)));
  if(loc>=(line+len)){
    *max=-1.0;
    *min=-1.0;
  }
  else{
    *max=strtod(loc,0);
  }
  return;
}


/*
 * Program entry point.
 */

int main(int argc, char **argv) {
    char serverpath[80];
    char *tempbuf,*tempbuf2;
    int ks_pick, ks_continue;
    poptContext optCon;
    int help=0;
    int listservers=0;
    int listmonitors=0;
    int listcards=0;
    int tmp;
    int dri = 0;
    int rc;
    int childpid, fd, status;
    char *nargv[10];
    int i;
    struct vt_stat vts;
        
    newtComponent f;
    enum { MAIN_SETUP, MAIN_TOP, MAIN_CARD, MAIN_MONITOR,
	   MAIN_SCREEN, MAIN_TEST, MAIN_DONE } step = MAIN_SETUP;

    struct poptOption cmdOptions[] = {
	{"pick", '\0', POPT_ARG_NONE, &pick, 0},
	{"continue", '\0', POPT_ARG_NONE, &rhcontinue, 0},
	{"kickstart", '\0', POPT_ARG_NONE, &kickstart, 0},
	{"card", '\0', POPT_ARG_STRING, &ks_card, 0},
	{"server", '\0', POPT_ARG_STRING, &ks_server, 0},
	{"driver", '\0', POPT_ARG_STRING, &ks_driver, 0},
	{"monitor", '\0', POPT_ARG_STRING, &ks_monitor, 0},
	{"hsync", '\0', POPT_ARG_STRING, &ks_hsync, 0},
	{"vsync", '\0', POPT_ARG_STRING, &ks_vsync, 0},
	{"bandwidth", '\0', POPT_ARG_STRING, &ks_bandwidth, 0},
	{"mouse", '\0', POPT_ARG_STRING, &ks_mouse, 0},
	{"mousedevice", '\0', POPT_ARG_STRING, &ks_mousedevice, 0 },
	{"emulthree", '\0', 0, &ks_emulate3, 0 },
	{"expert", '\0', POPT_ARG_NONE, &expert, 0},
	{"help", '\0', POPT_ARG_NONE, &help, 0},
	{"listservers", '\0', POPT_ARG_NONE, &listservers, 0},
	{"listmonitors", '\0', POPT_ARG_NONE, &listmonitors, 0},
	{"listcards", '\0', POPT_ARG_NONE, &listcards, 0},
	{"startxonboot", '\0', POPT_ARG_NONE, &onboot, 0},
	{"test", '\0', POPT_ARG_NONE, &testing, 0 },
	{"back", '\0', POPT_ARG_NONE, &back, 0 },
	{"preferxf3",'\0', POPT_ARG_NONE, &preferxf3,0},
	{"preferxf4",'\0', POPT_ARG_NONE, &preferxf4,0},
	{"nodri",'\0',POPT_ARG_NONE,&nodri,0},
	{"dri",'\0',POPT_ARG_NONE,&dri,0},
	{"noddcprobe",'\0',POPT_ARG_NONE, &noddcprobe,0},
	{"usefbdev",'\0',POPT_ARG_NONE,&usefbdev,0},
	{ 0, 0, 0, 0, 0 }
    };
    struct XconfigState myState, *state = &myState;

    if (parse_database()) {
	fprintf(stderr, _("\n\nCan't read card database.  Exiting.\n\n"));
	exit(1);
    }

    if((preferxf3==1)&&(preferxf4==1)){
        fprintf(stderr,_("\n\nCan't prefer both XFree 3.3.x and XFree 4.x "
                         "server.  Exiting.\n\n"));
        exit(1);
    }
        
    memset(state, 0, sizeof(myState));

    setlocale(LC_ALL, "");
    bindtextdomain("XorgTest","/usr/share/locale");
    textdomain("XorgTest");

    progName = argv[0];

    if (getuid() != 0) {
	fprintf(stderr, _("You must be root to run XorgTest...\n"));
	exit(1);
    }

    pick = 0;
    rhcontinue = 0;
    kickstart = 0;

    optCon = poptGetContext(NULL, argc,(const char**) argv, cmdOptions, 0);

    if ((rc = poptGetNextOpt(optCon)) < -1) {
	fprintf(stderr, "%s: %s\n",
		poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
		poptStrerror(rc));
	exit(1);
    }

    if ((dri==1)&&(nodri==1)) {
	fprintf(stderr, _("\n\nCan't both enable and disable DRI.  Exiting.\n\n"));
    }
    if (dri) {
	    nodri = 0;
    }

    if (onboot)
	state->runLevel = 4;
    else
	state->runLevel = 3;

    if ((tempbuf = (char *) poptGetArg(optCon))) {
	fprintf(stderr, _("unexpected argument %s -- use --help for usage "
                          "information\n"), tempbuf);
	exit(1);
    }

    poptFreeContext(optCon);

    if (help) usage(0);

    if (listservers) {
	ListServerTypes();
	return 0;
    }

    if (listmonitors) {
	ListMonitorTypes();
	return 0;
    }

    if (listcards) {
	ListCardTypes();
	return 0;
    }
    
    /* translate the arrays used in menus */
    translate_array(clockchip_name);
    translate_array(ramdac_name);
    translate_array(monitortype_name);

    normal = !pick && !rhcontinue && !kickstart;
    ks_pick = pick && kickstart;
    ks_continue = rhcontinue && kickstart;

    /* cant have expert and kickstart */
    if (kickstart)
	expert = 0;

    /* remove old /var/run/SERVER file so we dont get confused */
    if (!ks_continue && !rhcontinue && !access("/var/run/SERVER", F_OK))
	unlink("/var/run/SERVER");
    
    backupXConfigs();
    /* XXX - intl */
    while (step != MAIN_DONE)
	switch(step) {
	case MAIN_SETUP: /* Setup is to be called only once. */
	    if (rhcontinue)
		tempbuf = read_file();

	    if ((normal || pick || rhcontinue) && !kickstart) {
		newtInit();
		
		newtCls();
                tempbuf2=malloc(200);
                snprintf(tempbuf2,200,_("               XorgTest-%s  -  (C)2006  Gilbert Ashley and others"),VERSION);
                newtDrawRootText(0, 0, tempbuf2);
                newtPushHelpLine(_("  <Tab>/<Alt-Tab> move between elements | <Space> selects | <F12> next screen"));
                free(tempbuf2);
	    }
	    step = MAIN_TOP;
	    break;
	case MAIN_TOP:
	    if (pick || rhcontinue || kickstart) {
		step = MAIN_CARD;
		break;
	    }
	    if (newtWinChoice(_("Welcome"), _("Ok"),
			      back ? _("Back") : _("Cancel"),
			      _(intro_text)) == 2) {
		if (back) {
		    newtFinished();
		    return 1;
		} else
		    get_out();
	    }
	    step = MAIN_CARD;
	    break;
	case MAIN_CARD:
	    if (rhcontinue) {
		/* turn off expert temporarily */
		tmp = expert;
		expert = 0;
		if (carddb_configuration(state, tempbuf) == 1) {
		    step = MAIN_DONE;
		    break;
		}
/*  		rhcontinue=0; */
		expert = tmp;
		step = MAIN_MONITOR;
		break;
	    }

	    if (normal || pick || kickstart) {
		tmp = carddb_configuration(state, tempbuf);
		if (tmp < 0) {
		    fprintf(stderr,
			    _("No card determined in kickstart - bailing\n"));
		    exit(1);
		}
		if (tmp == 1) {
		    step = MAIN_TOP;
		    break;
		}
	    }

	    if (pick) {
		if (!kickstart) newtFinished();
		if (pick)
		    write_file(state);
		step = MAIN_DONE;
		break;
	    }
	case MAIN_MONITOR:
	    /* if (normal)
		XF86check_link(); */

	    chdir("/");
	    if (normal || rhcontinue || (kickstart && !pick) ) {
                if( (state->xf4driver == NULL)||( (state->serverbuf)&&(preferxf3)) ){
                    strcpy(serverpath, "../../usr/X11R6/bin/");
                    if(strncmp("XF86_",state->serverbuf,5)!=0)
                        strcat(serverpath,"XF86_");
		    strcat(serverpath, state->serverbuf);
                }
                else{
                    strcpy(serverpath,"../../usr/X11R6/bin/Xorg");
                }
                if (!access(serverpath, X_OK)) {
		    if (!testing) {
			unlink("/etc/X11/X");
#ifdef __sparc__
			if (!strncmp("xsun", state->serverbuf, 4)) {
			    /* Sun ignores XF86Config, but we need to
			       set the fontpath.  So we'll write a quick
			       wrapper here */
			    FILE *f;
			    f = fopen("/etc/X11/X", "w");
			    fprintf(f, "#!/bin/bash\n"
                                    "exec %s -fp unix/:7100 $@\n",
				    strstr(serverpath, "/usr"));
			    fclose(f);
			    chmod("/etc/X11/X", 0755);
			} else
#endif /* __sparc__ */
                            symlink(serverpath, "../../../etc/X11/X");
		    }
		} else {
		    if (!kickstart)
			newtFinished();
		    fprintf(stderr,
			    ("\n\tServer doesn't exist, can't continue.\n"));
		    fprintf(stderr, _("tried to use %s\n"), serverpath);
		    exit(1);
		}
		
		if (state->screentype <= 4)
		    mouse_configuration(state);
		
		keyboard_configuration(state);
	    }
	    step = MAIN_SCREEN;
	    break;
	case MAIN_SCREEN:
	    if (normal || rhcontinue || (kickstart && !pick)) {
		if (state->screentype <= 4) {
                    tmp=monitor_configuration(state);
		    if (tmp == 1) { 
			step = MAIN_TOP;
			break;
		    }
                    else if (tmp == 2){
                        step=MAIN_SCREEN;
                        break;
                    }
		    if (kickstart) {
			kickstart_screen_configuration(state);
		    } else {
                        if (screen_autoprobe(state) == 1) {
			    if (screen_configuration(state) == 1)
			        break;
                        }
		    }
		}
	    }
	    step = MAIN_TEST;
	    break;
	case MAIN_TEST:
	    if (!testing) write_XF86Config(state, ask_XF86location());
	    if (!kickstart && !onboot && !pick) {
		rc = newtWinChoice(_("Starting X"), _("Ok"), _("Skip"),
                                   _("XorgTest will now start X to test your "
                                     "configuration."));
		if (rc == 2) {
		    step = MAIN_DONE;
		    break;
		}
		newtCenteredWindow(60, 3, _("Starting X"));
		f = newtForm(NULL, NULL, 0);
		newtFormAddComponent(f, newtLabel(1, 1,
                                                  _("Preparing to start the X server...")));
		newtDrawForm(f);
		newtRefresh();
		newtFormDestroy(f);

		ioctl(0, VT_GETSTATE, &vts);
		
		if (!(childpid = fork())) {
		    if ((fd = open("/dev/null", O_WRONLY)) > 0)
			dup2(fd, 2);
		    
		    i = 0;
		    nargv[i++] = "/usr/sbin/Xtest";
		    if (normal)
			nargv[i++] = "--normal";
		    if (back)
			nargv[i++] = "--back";
		    if (slow_X())
			nargv[i++] = "--slow";
		    nargv[i++] = NULL;
		    execv(nargv[0], nargv);
		}

		waitpid(childpid, &status, 0);
		rc = 0;
		if (WIFEXITED(status) || WTERMSIG(status)) 
                    rc = WEXITSTATUS(status);

		/* unknown return code */
		if (rc && !(rc & ERROR) && !(rc & STARTX))
		    rc |= ERROR;

		newtPopWindow();
		newtRefresh();

		/* Switch back to the VC where we started. */
		ioctl(0, VT_ACTIVATE, vts.v_active);

		if (rc & ERROR) { /* error occured */
		    if (newtWinChoice(_("Error"), _("Back"), _("Quit"),
				      _("There is a problem with your X "
					"configuration.  You may go back "
					"and modify your configuration or "
					"exit now.")) == 2) {
			    newtFinished();
			    exit(1);

		    } else 
			step = MAIN_SCREEN;
		    
		    break;
		}

		if (rc & STARTX)
                    state->runLevel = 4;
		else
                    state->runLevel = 3;
		
		/* Success */
		rewriteInittab(state);
		startXfs(normal);
		newtFinished();
		_exit(0);
		break;
	    } else {
                if (kickstart && onboot)
		    rewriteInittab(state);
                step = MAIN_DONE;
	    }
        case MAIN_DONE:
            if( (state->xf4driver == NULL)||( (state->serverbuf)&&(preferxf3)) ){
                strcpy(serverpath, "../../usr/X11R6/bin/");
                if(strncmp("XF86_",state->serverbuf,5)!=0)
                    strcat(serverpath,"XF86_");
                strcat(serverpath, state->serverbuf);
            }
            else{
                strcpy(serverpath,"../../usr/X11R6/bin/Xorg");
            }

            startXfs(normal);
            break;
        }

    if (rhcontinue && !kickstart)
	newtFinished();

    if (normal) {
	newtCls();
	newtWinMessage(_("You're Done!"), _("Ok"), _(finalcomment_text));
	newtFinished();
    }

    exit(0);
}
