#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libintl.h>
#include <sys/wait.h>
#include <popt.h>
#include <locale.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "Xtest.h"

static char *finalcomment_text =
N_("Configuration file has been written. Take a look at it before starting X. Note that "
"the xorg.conf file must be in one of the directories searched by the server "
"(e.g. /etc/X11/xorg.conf) in order to be used. While the server is running, pressing "
"ctrl, alt and '+' simultaneously cycles through the video resolutions. Pressing ctrl, "
"alt and backspace simultaneously immediately exits the server (use if "
"the monitor doesn't sync for a particular mode). "
"\n\n"
"For further configuration, refer to /usr/X11R6/lib/X11/doc/README.Config.\n"
"\n");

static gboolean works, startx;
static int normal, back;
static int runlevel;
static int nostart = 0;
static int norunlevel = 0;
static int slow = 0;
int rc = 0;

#define DESKTOP_BG_COLOR  0x4A55AC

static char *initlevel_text =
N_("XorgTest can set up your computer to boot "
   "to a GUI login using runlevel 4.\n\n"
   "Would you like to setup a GUI login as the default? ");

static void destroy(GtkWidget *widget, gpointer data)
{
    gtk_main_quit ();
}

struct TimeoutData {
    GtkWidget *label;
    GtkWidget *window;
    gint time;
    gint timeout_id;
};

struct InitData {
  int runlevel;
  GtkWidget *window;
};

static void init_yes_callback (GtkWidget *widget, gpointer *data) {
    struct InitData * init_data = (struct InitData *) data;

    init_data->runlevel = 4;
    startx = 1;
    gtk_widget_destroy (init_data->window);
    g_free(init_data);
    gtk_main_quit ();
}

static void init_no_callback (GtkWidget *widget, gpointer *data) {
    struct InitData * init_data = (struct InitData *) data;

    init_data->runlevel = 3;
    gtk_widget_destroy (init_data->window);
    g_free(init_data);
    gtk_main_quit ();
}

static void yes_callback (GtkWidget *widget, gpointer *data) {
    struct TimeoutData *timeout_data = (struct TimeoutData *) data;

    works = 1;
    gtk_widget_destroy (timeout_data->window);
    gtk_timeout_remove(timeout_data->timeout_id);
    g_free(timeout_data);
    gtk_main_quit ();
}

static void no_callback (GtkWidget *widget, gpointer *data) {
    struct TimeoutData *timeout_data = (struct TimeoutData *) data;

    gtk_widget_destroy (timeout_data->window);
    gtk_timeout_remove(timeout_data->timeout_id);
    g_free(timeout_data);
    gtk_main_quit ();
}

static gint timeout_callback(gpointer data)
{
    char buf[256];

    struct TimeoutData *timeout_data = (struct TimeoutData *) data;

    timeout_data->time--;

    snprintf(buf, sizeof(buf), _("If you can see this message, click \'YES\'\n\n"
	                         "Automatic timeout in: %d seconds"),
	     timeout_data->time);
    gtk_label_set_text(GTK_LABEL(timeout_data->label), buf);

    if (timeout_data->time == 0) {
	gtk_widget_destroy (timeout_data->window);
	gtk_timeout_remove(timeout_data->timeout_id);
	g_free(timeout_data);
	gtk_main_quit ();
	return FALSE;
    }
    return TRUE;
}

static GtkWidget *accel_button(GtkAccelGroup *group, const char *text)
{
    GtkWidget *button, *label;
    gchar key;
    
    label = gtk_label_new(text);
    key = gtk_label_parse_uline(GTK_LABEL(label), text);
    button = gtk_button_new();
    gtk_container_add (GTK_CONTAINER(button), label);
    gtk_widget_add_accelerator (button, "clicked", group, key, 0, 0);

    return button;
}
     
static void create_test_window(void)
{
    GtkWidget *window, *label, *button, *vbox, *hbox, *frame;
    GtkAccelGroup *accelgroup;
    char buf[256];
    struct TimeoutData *timeout_data;

    timeout_data = g_malloc(sizeof(struct TimeoutData));

    window = gtk_window_new (GTK_WINDOW_POPUP);

    accelgroup = gtk_accel_group_new();
    gtk_accel_group_attach(accelgroup, GTK_OBJECT(window));

    vbox = gtk_vbox_new (FALSE, 0);
    frame = gtk_frame_new (NULL);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
    gtk_container_border_width (GTK_CONTAINER (vbox), 25);

    gtk_container_add (GTK_CONTAINER (frame), vbox);
    gtk_container_add (GTK_CONTAINER (window), frame);

    snprintf(buf, sizeof(buf), _("Testing graphics mode %dx%d pixels with %d bits per pixel\n"),
	     gdk_screen_width(), gdk_screen_height(),
	     DefaultDepthOfScreen (DefaultScreenOfDisplay (GDK_DISPLAY())));
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);

    snprintf(buf, sizeof(buf), _("If you can you see this message, click \'YES\'\n\n"
	                       "Automatic timeout in: %d seconds"), slow ? 60 : 10);

    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
    timeout_data->time = slow ? 60 : 10;
    timeout_data->label = label;
    timeout_data->window = window;

    timeout_data->timeout_id =
	gtk_timeout_add(1000, timeout_callback, timeout_data);

    hbox = gtk_hbox_new (TRUE, 10);
    button = accel_button(accelgroup, _("_Yes"));
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC(yes_callback),
			timeout_data);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);

    gtk_window_set_focus(GTK_WINDOW(window), button);

    button = accel_button(accelgroup, _("_No"));
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC(no_callback),
			timeout_data);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);

    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

    gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			GTK_SIGNAL_FUNC (destroy), window);

    gtk_signal_connect (GTK_OBJECT (window), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);


    gtk_window_set_position (GTK_WINDOW (window),
			     GTK_WIN_POS_CENTER);
    
    gtk_widget_show_all (window);

    /* default is the X cursor.  We would like an arrow pointer. */
    gdk_window_set_cursor(window->window, gdk_cursor_new(GDK_LEFT_PTR));
    
/*    gdk_keyboard_grab (window->window, FALSE, GDK_CURRENT_TIME); */
}

static GdkPixbuf *open_sample_image(gchar *file) {
    GdkPixbuf *retval = NULL;
    gchar *path;

    path = g_strdup_printf("%s/%s", SAMPLESCREENS_DIR, file);
    retval = gdk_pixbuf_new_from_file (path);
    g_free(path);

    return retval;
}

static void set_background(void)
{
    gint output_width, output_height;
    gint rt_width, rt_height, lt_width, lt_height, mid_width, mid_height;
    GdkPixbuf *panel_left  = NULL;
    GdkPixbuf *panel_right = NULL;
    GdkPixbuf *panel_mid   = NULL;
    GdkPixbuf *desktop_icons = NULL;
    GdkPixbuf *panel_mid_rendered;
    GdkPixmap *bg_pixmap;
    GdkGC *gc;
    GdkColor bgcolor;

    gdk_rgb_init();

    output_width = gdk_screen_width ();
    output_height = gdk_screen_height ();

    bg_pixmap = GDK_ROOT_PARENT();
    gc = gdk_gc_new (bg_pixmap);
    gdk_rgb_gc_set_foreground (gc, DESKTOP_BG_COLOR);

/*    gdk_draw_rectangle (bg_pixmap, gc, TRUE, 
			0, 0, output_width, output_height);
*/
    bgcolor.pixel = gdk_rgb_xpixel_from_rgb ( DESKTOP_BG_COLOR );
    gdk_window_set_background(bg_pixmap, &bgcolor);
    XClearWindow (GDK_DISPLAY(), GDK_ROOT_WINDOW());

    panel_left = open_sample_image ("panel_left.png");
    panel_right = open_sample_image ("panel_right.png");
    panel_mid   = open_sample_image ("panel_mid.png");
    desktop_icons = open_sample_image ("desktop_icons.png");

    if (panel_left && panel_right && panel_mid && desktop_icons) {
	lt_width  = gdk_pixbuf_get_width (panel_left);
	lt_height = gdk_pixbuf_get_height (panel_left);
	gdk_pixbuf_render_to_drawable (panel_left, bg_pixmap, gc, 
				       0, 0, 
				       0, output_height - lt_height, 
				       lt_width, lt_height,
				       GDK_RGB_DITHER_MAX, 0, 0);

	rt_width  = gdk_pixbuf_get_width (panel_right);
	rt_height = gdk_pixbuf_get_height (panel_right);
	gdk_pixbuf_render_to_drawable (panel_right, bg_pixmap, gc, 
				       0, 0, 
				       output_width - rt_width, 
				       output_height - rt_height, 
				       rt_width, rt_height,
				       GDK_RGB_DITHER_MAX, 0, 0);
	/* assume all the same height */
	mid_width  = output_width  - rt_width - lt_width;
	mid_height = rt_height;
	panel_mid_rendered = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
					     FALSE, 8, mid_width, mid_height);

        gdk_pixbuf_scale (panel_mid, panel_mid_rendered,
			  0, 0, mid_width, mid_height, 0, 0,
			  (double)mid_width, 1.0, GDK_INTERP_TILES);

	gdk_pixbuf_render_to_drawable (panel_mid_rendered, bg_pixmap, gc, 
				       0, 0, 
				       lt_width, output_height - mid_height, 
				       mid_width, mid_height,
				       GDK_RGB_DITHER_MAX, 0, 0);

	gdk_pixbuf_render_to_drawable_alpha (desktop_icons, bg_pixmap, 
				       0, 0, 0, 0,
				       gdk_pixbuf_get_width (desktop_icons), 
				       gdk_pixbuf_get_height (desktop_icons),
 				       GDK_PIXBUF_ALPHA_BILEVEL,
				       255,
				       GDK_RGB_DITHER_MAX, 0, 0);


	/* free everything */
	gdk_pixbuf_unref(panel_right);
	gdk_pixbuf_unref(panel_mid);
	gdk_pixbuf_unref(panel_mid_rendered);
	gdk_pixbuf_unref(panel_left);
	gdk_pixbuf_unref(desktop_icons);
    }
}


static void create_final_window(void)
{
    GtkWidget *window, *label, *button, *vbox, *frame;
    GtkAccelGroup *accelgroup;
    
    window = gtk_window_new (GTK_WINDOW_POPUP);
    frame = gtk_frame_new (NULL);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
    
    accelgroup = gtk_accel_group_new();
    gtk_accel_group_attach(accelgroup, GTK_OBJECT(window));

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_border_width (GTK_CONTAINER (vbox), 5);
    gtk_container_add (GTK_CONTAINER (frame), vbox);
    gtk_container_add (GTK_CONTAINER (window), frame);

    label = gtk_label_new(_(finalcomment_text));
    gtk_label_set_line_wrap (GTK_LABEL(label), TRUE);

    gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);

    gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			GTK_SIGNAL_FUNC (destroy), window);

    gtk_signal_connect (GTK_OBJECT (window), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);

    button = accel_button(accelgroup, _("_OK"));
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC(destroy), NULL);
    gtk_window_set_focus(GTK_WINDOW(window), button);
    gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 0);

    gtk_window_set_position (GTK_WINDOW (window),
			     GTK_WIN_POS_CENTER);

    gtk_widget_show_all (window);

    gdk_window_set_cursor(window->window, gdk_cursor_new(GDK_LEFT_PTR));

}

static void create_initlevel_window(int runlevel)
{
    GtkWidget *window, *label, *button, *vbox, *hbox, *frame;
    struct InitData *init_data;
    GtkAccelGroup *accelgroup;
    
    init_data = g_malloc(sizeof(struct InitData));

    window = gtk_window_new (GTK_WINDOW_POPUP);

    frame = gtk_frame_new (NULL);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_border_width (GTK_CONTAINER (vbox), 5);
    
    gtk_container_add (GTK_CONTAINER (frame), vbox);
    gtk_container_add (GTK_CONTAINER (window), frame);
    
    accelgroup = gtk_accel_group_new();
    gtk_accel_group_attach(accelgroup, GTK_OBJECT(window));

    label = gtk_label_new(_(initlevel_text));
    gtk_label_set_line_wrap (GTK_LABEL(label), TRUE);

    gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);

    gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			GTK_SIGNAL_FUNC (destroy), window);

    gtk_signal_connect (GTK_OBJECT (window), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);

    hbox = gtk_hbox_new (TRUE, 10);

    init_data->window = window;
    init_data->runlevel = runlevel;

    button = accel_button(accelgroup, _("_Yes"));
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC(init_yes_callback), init_data);

    gtk_window_set_focus(GTK_WINDOW(window), button);

    button = accel_button(accelgroup, _("_No"));
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC(init_no_callback), init_data);

    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

    gtk_window_set_position (GTK_WINDOW (window),
			     GTK_WIN_POS_CENTER);
    
    gtk_widget_show_all (window);

    gdk_window_set_cursor(window->window, gdk_cursor_new(GDK_LEFT_PTR));    

    gdk_keyboard_grab (window->window, FALSE, GDK_CURRENT_TIME);
}

int Xtest (int argc, char **argv)
{
    int childpid, fd;
    int status=0;
    int testpid;   
    
    works = startx = 0;

    if (!nostart) {
	if (!(childpid = fork())) {
	    if ((fd = open("/dev/null", O_WRONLY)) > 0);
	    dup2(fd, 2);
	    execl("/usr/X11R6/bin/X", "X", "-fp",
		  "/usr/X11R6/lib/X11/fonts/75dpi,"
		  "/usr/X11R6/lib/X11/fonts/misc,"
		  "/usr/share/fonts/ISO8859-2/75dpi,"
		  "/usr/share/fonts/ISO8859-2/misc,"
		  "/usr/share/fonts/ISO8859-7/75dpi,"
		  "/usr/share/fonts/ISO8859-7/misc,"
		  "/usr/share/fonts/ISO8859-9/75dpi,"
		  "/usr/share/fonts/ISO8859-9/misc,"
		  "/usr/X11R6/lib/X11/fonts/cyrillic",
		  ":9", NULL);
	}
	setenv("DISPLAY", ":9", 1);
    }
    sleep(3);

    if (!nostart) {
	testpid=waitpid(childpid, &status, WNOHANG);
 	if (testpid>0 && WIFEXITED(status)) 
            return ERROR; 
    }

    gtk_set_locale ();
    gtk_init (&argc, &argv);
    
    set_background();
    create_test_window();

    gtk_main();

    if (!works && !nostart) {
      kill(childpid, SIGTERM);
      return ERROR;
    }

    if (!norunlevel) {
	create_initlevel_window(runlevel);
	gtk_main();
    }

    if (normal) {
      create_final_window();
      gtk_main();
    }

    if (!nostart) {
	kill(childpid, SIGTERM);
    }
    
    if (startx)
      return STARTX;
    else
      return 0;
}

int main(int argc, char *argv[])
{
  poptContext optCon;
  struct poptOption cmdOptions[] = {
    {"back", 'b', POPT_ARG_NONE, &back, 0},
    {"normal", 'n', POPT_ARG_NONE, &normal, 0},
    {"runlevel", 'r', POPT_ARG_INT, &runlevel, 0},
    {"nostart", 's', POPT_ARG_NONE, &nostart, 0},
    {"norunlevel", 'l', POPT_ARG_NONE, &norunlevel, 0},
    {"slow", 'S', POPT_ARG_NONE, &slow, 0},
    POPT_AUTOHELP
    {0, 0, 0, 0, 0}
  };
  int retcode;

  setlocale(LC_ALL, "");
  bindtextdomain("XorgTest","/usr/share/locale");
  textdomain("XorgTest");

  normal = back = 0;
  runlevel = 3;

  optCon = poptGetContext(NULL, argc, (const char **)argv, cmdOptions, 0);

  if ((retcode = poptGetNextOpt(optCon)) < -1) {
    fprintf(stderr, "%s: %s\n",
	    poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
	    poptStrerror(retcode));
    _exit(ERROR);
  }

  if (poptGetArg(optCon)) {
    fprintf(stderr, "unexpected argument. Only --back, --normal, and "
	    "--runlevel are valid\n");
    _exit(ERROR);
  }
  
  poptFreeContext(optCon);

  rc = Xtest(argc, argv);
  
  _exit(rc);
}
