/* $XConsortium: cards.c,v 1.4 95/06/09 20:57:00 gildea Exp $ */
/* $XFree86: xc/programs/Xserver/hw/xfree86/xf86config/cards.c,v 3.5 1995/07/21 14:42:05 dawes Exp $ */

/*
 *  Functions to manipulate card database.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "cards.h"

/*
 * Database format:
 *
 * NAME <name of card>
 * CHIPSET <chipset description>
 * SERVER <server name>
 * DRIVER <driver name for XFree 4>
 *
 * You need at least one of DRIVER and SERVER
 *
 * Optional keywords:
 * RAMDAC <ramdac identifier>
 * CLOCKCHIP <clockchip identifier>
 * DACSPEED <dacspeed>
 * NOCLOCKPROBE
 * UNSUPPORTED
 *
 * SEE <name of card> refers to another card definition; parameters that
 * are already defined are not overridden.
 *
 * <server name> is one of Mono, VGA16, SVGA, S3, Mach32, Mach8, 8514,
 * P9000, AGX, W32.
 *
 * A useful additional keywords may be CLOCKS.
 */



/* Database vars. */

int lastcard;

Card card[MAX_CARDS];

void sort_database ();

static void
appendstring (destp, src)
     char **destp;
     char *src;
{
  char *newstr;
  newstr = malloc (strlen (*destp) + strlen (src) + 1);
  strcpy (newstr, *destp);
  strcat (newstr, src);
  if (strlen (*destp) > 0)
    free (*destp);
  *destp = newstr;
}

int
lookupcard (char *name)
{
  int i;
  for (i = 0; i <= lastcard; i++)
    if (strcmp (name, card[i].name) == 0)
      return i;
  return -1;
}

static char *s3_comment =
  "# Use Option \"nolinear\" if the server doesn't start up correctly\n"
  "# (this avoids the linear framebuffer probe). If that fails try\n"
  "# option \"nomemaccess\".\n"
  "#\n"
  "# Refer to /usr/X11R6/lib/doc/README.S3, and the XF86_S3 man page.\n";

static char *cirrus_comment =
  "# Use Option \"no_bitblt\" if you have graphics problems. If that fails\n"
  "# try Option \"noaccel\".\n"
  "# Refer to /usr/X11R6/lib/doc/README.cirrus.\n"
  "# To allow linear addressing, uncomment the Option line and the\n"
  "# address that the card maps the framebuffer to.\n";

int
parse_database ()
{
  FILE *f;
  char buf[256];
  int i, lineno, errors;
  f = fopen (CARD_DATABASE_FILE, "r");
  if (f == NULL)
    {
#ifdef __sparc__
      lastcard = -1;
      return 0;
#else
      return -1;
#endif
    }

  lastcard = -1;
  lineno = 0;
  errors = 0;

  for (;;)
    {
      if (fgets (buf, 256, f) == NULL)
	/* be sure we visit validity check */
	strcpy (buf, "END");
      lineno++;
      if (buf[0] == '#')
	/* Comment. */
	continue;

      /* If at the end of an entry... check for validity */
      if (lastcard > 0 && !(strncmp (buf, "NAME", 4)
			    && strncmp (buf, "END", 3)))
	{
	  /* test for missing required fields */
            if ((card[lastcard].server == NULL) && (card[lastcard].driver == NULL)){
                fprintf (stderr, "\rWarning - no SERVER or DRIVER specification"
                         " in Card database entry %s"
                         " (line %d).\n", card[lastcard].name, lineno);
                card[lastcard].server = "unknown";
                errors++;
	    }

/* We can do without this..... */

/*  	  if (card[lastcard].chipset == NULL) */
/*  	    { */
/*  	      fprintf (stderr, "\rWarning CHIPSET specifica" */
/*  		       "tion missing in Card database entry %s" */
/*  		       " (line %d).\n", card[lastcard].name, lineno); */
/*  	      card[lastcard].chipset = "unknown"; */
/*  	      errors++; */
/*  	    } */
	}
      if (strncmp (buf, "END", 3) == 0)
	/* End of database. */
	break;
      if (strncmp (buf, "LINE", 4) == 0 && lastcard >= 0)
	{
	  /* If there is an uncommented VideoRam LINE in the Cards
	     database, ignore it. */
	  if ((*(buf + 5) != '#') && !(strncasecmp (buf + 5, "VideoRam", 8)))
	    continue;
	  /* Line of Device comment. */
	  /* Append to existing lines. */
	  appendstring (&card[lastcard].lines, buf + 5);
	  continue;
	}
      /*
       * The following keywords require the trailing newline
       * to be deleted.
       */
      i = strlen (buf);
      buf[--i] = '\0';

      /* remove trailing spaces or tabs */
      for (--i; i >= 0 && (buf[i] == ' ' || buf[i] == '\011'); i--);
      if (i >= 0)
	buf[i + 1] = '\0';
      else
	continue;		/* skip empty lines */

      if (strncmp (buf, "NAME", 4) == 0)
	{
	  /* New entry. */
	  lastcard++;
	  card[lastcard].name = malloc (strlen (buf + 5) + 1);
	  strcpy (card[lastcard].name, buf + 5);
	  card[lastcard].chipset = NULL;
	  card[lastcard].ramdac = NULL;
	  card[lastcard].clockchip = NULL;
	  card[lastcard].dacspeed = NULL;
          card[lastcard].driver =NULL;
	  card[lastcard].flags = 0;
	  card[lastcard].lines = "";
	  continue;
	}
      if (lastcard < 0)		/* no NAME line found yet */
	continue;
      if (strncmp (buf, "SEE", 3) == 0)
	{
	  /* Reference to another entry. */
	  int i;
	  i = lookupcard (buf + 4);
	  if (i == -1)
	    {
	      fprintf (stderr, "\rError in database, invalid"
		       " reference: %s.\n", buf + 4);
	      errors++;
	      continue;
	    }
	  if (card[lastcard].chipset == NULL)
              card[lastcard].chipset = card[i].chipset;
	  if (card[lastcard].server == NULL)
              card[lastcard].server = card[i].server;
	  if (card[lastcard].ramdac == NULL)
              card[lastcard].ramdac = card[i].ramdac;
	  if (card[lastcard].clockchip == NULL)
              card[lastcard].clockchip = card[i].clockchip;
	  if (card[lastcard].dacspeed == NULL)
              card[lastcard].dacspeed = card[i].dacspeed;
	  if (card[lastcard].driver == NULL)
              card[lastcard].driver = card[i].driver;
	  card[lastcard].flags |= card[i].flags;
	  appendstring (&card[lastcard].lines, card[i].lines);
	  continue;
	}
      if (strncmp (buf, "CHIPSET", 7) == 0)
	{
	  /* Chipset description. */
	  card[lastcard].chipset = malloc (strlen (buf + 8) + 1);
	  strcpy (card[lastcard].chipset, buf + 8);
	  continue;
	}
      if (strncmp (buf, "SERVER", 6) == 0)
	{
	  /* Server identifier. */
	  card[lastcard].server = malloc (strlen (buf + 7) + 1);
	  strcpy (card[lastcard].server, buf + 7);
	  continue;
	}
      if (strncmp (buf, "RAMDAC", 6) == 0)
	{
	  /* Ramdac indentifier. */
	  card[lastcard].ramdac = malloc (strlen (buf + 7) + 1);
	  strcpy (card[lastcard].ramdac, buf + 7);
	  continue;
	}
      if (strncmp (buf, "CLOCKCHIP", 9) == 0)
	{
	  /* Clockchip indentifier. */
	  card[lastcard].clockchip = malloc (strlen (buf + 10) + 1);
	  strcpy (card[lastcard].clockchip, buf + 10);
	  card[lastcard].flags |= NOCLOCKPROBE;
	  continue;
	}
      if (strncmp (buf, "DRIVER", 6) == 0){
              /* XFree 4 driver */
          card[lastcard].driver = malloc (strlen (buf + 7) + 1);
          strcpy (card[lastcard].driver, buf + 7);
          continue;
      }
      if (strncmp (buf, "DACSPEED", 8) == 0)
	{
	  /* Clockchip indentifier. */
	  card[lastcard].dacspeed = malloc (strlen (buf + 9) + 1);
	  strcpy (card[lastcard].dacspeed, buf + 9);
	  continue;
	}
      if (strncmp (buf, "NOCLOCKPROBE", 12) == 0)
	{
	  card[lastcard].flags |= NOCLOCKPROBE;
	  continue;
	}
      if (strncmp (buf, "UNSUPPORTED", 11) == 0)
	{
	  card[lastcard].flags |= UNSUPPORTED;
	  continue;
	}
    }

  fclose (f);

  /*
   * Add general comments.
   */
  for (i = 0; i <= lastcard; i++){
      if(card[i].server)
          if (strcmp (card[i].server, "S3") == 0)
              appendstring (&card[i].lines, s3_comment);
      if(card[i].chipset)
          if (strncmp (card[i].chipset, "CL-GD", 5) == 0)
              appendstring (&card[i].lines, cirrus_comment);
  }
  
  sort_database ();

  /* if there were errors */
  if (errors)
    /* sleep for up to 5 secs so that the user can read them. */
    sleep (errors < 5 ? errors : 5);

  return 0;
}

#ifdef __STDC__
#define CONST const
#else
#define CONST
#endif

static int
compare_card (e1, e2)
     CONST void *e1;
     CONST void *e2;
{
  return strcmp (((Card *) e1)->name, ((Card *) e2)->name);
}

void
sort_database ()
{
  /* Each element is a bunch of words, but nothing too bad. */
  qsort (card, lastcard + 1, sizeof (Card), compare_card);
}
