#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "combo.h"
#include "XorgTest.h"

/*
 * Documentation in combo.h
 */

GSList* loadComboFile(char *filename){
  
    FILE *in;
    GSList *res;
    char line[256]; /* Max length of line in file */
    char *pos;
    void *chk;
    int id,depth;
    struct CardMonitorCombo *combo;
    char *resolutions;
    
    res=NULL;
    in=fopen(filename,"r");
    if(in==NULL)
        return NULL;
    
    while(fgets(line,256,in)!=NULL){
        line[strlen(line)-1]=0;
        if((line[0]=='#')||(line[0]=='\0'))
            continue;
        if(sscanf(line,"PCIID %i",&id)==0)
            continue;
        else
            combo=calloc(sizeof(struct CardMonitorCombo),1);
        combo->pciid=id;
        chk=fgets(line,256,in);
        sscanf(line,"VENDORID %i",&(combo->pcivendorid));
        chk=fgets(line,256,in);
        line[strlen(line)-1]=0;
        if((pos=strcasestr(line,"MONITORID "))==0){
            continue;
        };
        combo->monitorid=strdup(&line[10]);
        chk=fgets(line,256,in);
        if(chk==NULL)
            continue;
        line[strlen(line)-1]=0;
        while(strcasestr(line,"modeline ")!=0){
            combo->modelines=g_slist_prepend(combo->modelines, strdup(line));
            chk=fgets(line,256,in);
            if(chk==NULL){
                break;
            }
            line[strlen(line)-1]=0;
        }
        while(strcasestr(line,"RESOLUTION")!=0){
            resolutions=treatResLine(line+10,&depth);
            switch(depth){
            case 8:
                combo->modes8bit=resolutions;
                break;
            case 16:
                combo->modes16bit=resolutions;
                break;
            case 32:
                combo->modes32bit=resolutions;
                break;
            default:
                free(resolutions);
                break;
            }
            chk=fgets(line,256,in);
            if(chk==NULL){
                break;
            }
            line[strlen(line)-1]=0;
        }
        res=g_slist_append(res,combo); 
    }
    fclose(in);
    return res;
}

struct CardMonitorCombo* findCombo(GSList *combolist,
                                   struct XconfigState* state)
{
    GSList *tmp;
    struct CardMonitorCombo *combo;
    
    if( state->pciid==0 || state->pcivendorid==0  || combolist==NULL){
        return NULL;
    }
    tmp=combolist;
    while(tmp!=NULL){
        combo=tmp->data;
        if( (state->pciid==combo->pciid) &&
            (state->pcivendorid==combo->pcivendorid) &&
            !strcmp(state->monitoridentifier,combo->monitorid)
	   ){
		return combo;
        }
        tmp=g_slist_next(tmp);
    };
    return NULL;
}

char* treatResLine(char* line,int *depth)
{
/*      gchar** splitres; */
    char *temp;
    char *res;
    
    /* Max 8 resolutions.... */

    res=calloc(sizeof(char*),2);
    *depth=atoi(line);
    temp=strchr(line,' ')+1;
    
/*      splitres=g_strsplit(temp," ",8); */
    res=strdup(temp);
    return res;
}

void freeComboList(GSList *list){
    GSList *tmp,*tmp2;
    struct CardMonitorCombo *combo;

    tmp=list;
    while(tmp!=NULL){
        combo=tmp->data;
        free(combo->monitorid);
        if(combo->modes8bit)
            free(combo->modes8bit);
        if(combo->modes16bit)
            free(combo->modes16bit);
        if(combo->modes32bit)
            free(combo->modes32bit);
        tmp2=combo->modelines;
        while(tmp2!=NULL){
            free(tmp2->data);
            tmp2=g_slist_next(tmp2);
        }
        g_slist_free(tmp2);
        tmp=g_slist_next(tmp);
    };
    g_slist_free(list);
}

    
char*  get8bitModes(struct CardMonitorCombo* combo){
    return combo->modes8bit;
}

    
char* get16bitModes(struct CardMonitorCombo* combo){
    return combo->modes16bit;
}

char* get32bitModes(struct CardMonitorCombo* combo){
    return combo->modes32bit;
}
